from numbers import Number

import pytest

from .common import (
    function_builder,
    multimethod_dispatch,
    multipledispatch_dispatch,
    ovld_dispatch,
    plum_dispatch,
    runtype_dispatch,
    singledispatch_dispatch,
    with_functions,
)


class Animal:
    pass


class Mammal(Animal):
    pass


class Cat(Mammal):
    pass


class Dog(Mammal):
    pass


class Bird(Animal):
    pass


###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_trivial(dispatch):
    @dispatch
    def trivial(x: Number):
        return "A"

    @dispatch
    def trivial(x: str):
        return "B"

    @dispatch
    def trivial(x: dict):
        return "C"

    @dispatch
    def trivial(x: list):
        return "D"

    @dispatch
    def trivial(x: Cat):
        return "E"

    @dispatch
    def trivial(x: Mammal):
        return "F"

    @dispatch
    def trivial(x: Animal):
        return "G"

    return trivial


##############
# isinstance #
##############


def trivial_isinstance(x):
    if isinstance(x, int | float):
        return "A"
    if isinstance(x, str):
        return "B"
    if isinstance(x, dict):
        return "C"
    if isinstance(x, list):
        return "D"
    if isinstance(x, Cat):
        return "E"
    if isinstance(x, Mammal):
        return "F"
    if isinstance(x, Animal):
        return "G"


####################
# Test definitions #
####################


@pytest.mark.benchmark(group="trivial")
@with_functions(
    ovld=make_trivial(ovld_dispatch),
    plum=make_trivial(plum_dispatch),
    multimethod=make_trivial(multimethod_dispatch),
    multipledispatch=make_trivial(multipledispatch_dispatch),
    runtype=make_trivial(runtype_dispatch),
    singledispatch=make_trivial(singledispatch_dispatch),
    custom__isinstance=trivial_isinstance,
)
def test_trivial(fn, benchmark):
    def run():
        return [
            fn(1),
            fn(3.5),
            fn("hello"),
            fn({}),
            fn([1, 2]),
            fn(Cat()),
            fn(Dog()),
            fn(Bird()),
        ]

    result = benchmark(run)
    assert result == list("AABCDEFG")



import inspect
import sys
from dataclasses import dataclass, fields
from itertools import count

import pytest

from ovld import codegen, recurse
from ovld.codegen import Code, Def, Lambda, code_generator
from ovld.core import OvldBase, ovld
from ovld.dependent import Regexp
from ovld.types import All, Dataclass
from ovld.utils import MISSING, CodegenInProgress


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Person:
    name: str
    hometown: str
    age: int


SEP = """
==========
"""


def test_code_gensym():
    codegen._current = count()
    co = Code("$=x if $:x else $y", x=Code("f(1, 2, 3)"), y=123)
    assert co.fill() == "x__0 if (x__0 := f(1, 2, 3)) else 123"


def test_code_gensym_simple():
    codegen._current = count()
    obj = object()
    co = Code("$=x if $:x else $y", x=obj, y=123)
    assert co.fill() == "x if x else 123"


def test_rename():
    lbda = Lambda(["x"], "$x + 1")
    expr = lbda(Code("($x * 2)"))

    assert expr.sub(x=123).fill() == "(123 * 2) + 1"


def test_rename_closure():
    lbda = Lambda(["x"], "$x + $y + $z", y=Code("$x", x=5), z=Code("$x"))
    expr = lbda(20)

    assert expr.fill() == "20 + 5 + 20"


def getcodes(fn, *sigs):
    sigs = [(sig if isinstance(sig, tuple) else (sig,)) for sig in sigs]
    codes = [inspect.getsource(fn.resolve(*sig)) for sig in sigs]
    return SEP.join(codes)


def test_simple(file_regression):
    @ovld
    @code_generator
    def f(x: object):
        return Code("return $cls", cls=x)

    assert f(int) is type
    assert f(123) is int
    assert f("wow") is str
    assert f(Point(1, 2)) is Point
    assert f("zazz") is str
    assert f(123) is int

    file_regression.check(getcodes(f, type, int, str, Point))


def test_dataclass_gen(file_regression):
    @ovld
    @code_generator
    def f(x: Dataclass):
        body = [f"{fld.name}=$recurse(x.{fld.name})," for fld in fields(x)]
        return Code(["return $cons(", body, ")"], cons=x, recurse=recurse)

    @ovld
    def f(x: int):
        return x + 1

    @ovld
    def f(x: str):
        return x[1:].capitalize()

    assert f(Point(1, 2)) == Point(2, 3)
    assert f(Person("Robert", "Montreal", 21)) == Person("Obert", "Ontreal", 22)

    file_regression.check(getcodes(f, Point, Person))


def test_method(file_regression):
    class Plusser:
        def __init__(self, number):
            self.number = number

        @ovld
        @code_generator
        def f(self, thing: object):
            assert self is MISSING
            return Code("return thing + $cls(self.number)", cls=thing)

    plus = Plusser(5)
    assert plus.f(3) == 8
    assert plus.f("wow") == "wow5"

    file_regression.check(getcodes(Plusser.f, int, str))


def test_method_metaclass(file_regression):
    class Plusser(OvldBase):
        def __init__(self, number):
            self.number = number

        @ovld
        @code_generator
        def f(self, thing: object):
            assert self is Plusser
            return Code("return thing + $cls(self.number)", cls=thing)

    plus = Plusser(5)
    assert plus.f(3) == 8
    assert plus.f("wow") == "wow5"

    file_regression.check(getcodes(Plusser.f, int, str))


def test_variant_generation(file_regression):
    def normal(fn):
        fn.normal = True
        return fn

    @ovld
    @code_generator
    def f(obj: Dataclass):
        lines = ["return $cons("]
        for fld in fields(obj):
            existing = recurse.resolve(fld.type)
            if getattr(existing, "normal", False):
                expr = f"obj.{fld.name}"
            else:
                expr = f"$recurse(obj.{fld.name})"
            lines.append(f"    {fld.name}={expr},")
        lines.append(")")
        return Code("\n".join(lines), cons=obj, recurse=recurse)

    @ovld
    @normal
    def f(obj: object):
        return obj

    @f.variant
    def g(obj: int):
        return obj + 1

    file_regression.check(getcodes(f, Person) + SEP + getcodes(g, Person))


def test_nogen():
    @ovld
    @code_generator
    def f(x: Dataclass):
        if any(fld.type is not int for fld in fields(x)):
            return None
        body = [f"{fld.name}=$recurse(x.{fld.name})," for fld in fields(x)]
        return Code(["return $cons(", body, ")"], cons=x, recurse=recurse)

    @ovld
    def f(x: int):
        return x + 1

    @ovld
    def f(x: object):
        return False

    assert f(Point(1, 2)) == Point(2, 3)
    assert f(Person("Robert", "Montreal", 21)) is False


def test_dependent_generation():
    @ovld
    @code_generator
    def f(obj: Regexp[r"^A"]):
        return "return 'rx'"

    @ovld
    def f(obj: str):
        return "s"

    assert f("Allo") == "rx"
    assert f("Banana") == "s"


def test_generate_function_directly():
    @ovld
    @code_generator
    def f(x: object):
        xt = x
        return lambda x: (xt, x)

    assert f("coconut") == (str, "coconut")


@dataclass
class Tree:
    left: "Tree | int"
    right: "Tree | int"


@dataclass
class Point:
    x: int
    y: int


@dataclass
class TwoPoints:
    a: Point
    b: Point


@dataclass
class Empty:
    pass


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10+ for UnionType")
def test_inlining_generator():
    from types import UnionType

    def resolve_subcode(t, f):
        try:
            fn = f.resolve(type[t], All)
            lbda = getattr(fn, "__codegen__", None)
            if lbda:
                return lbda
        except CodegenInProgress:
            pass
        return Lambda(["typ", "x"], Code("$recurse($typ, $x)", recurse=f))

    @ovld
    @code_generator
    def f(typ: type[int], x: int):
        return Lambda(["typ", "x"], "$x + 1")

    @ovld
    @code_generator
    def f(typ: type[UnionType], x: object):
        (typ,) = typ.__args__
        subcodes = [(t, resolve_subcode(t, recurse)) for t in typ.__args__]
        assert len(subcodes) == 2
        return Lambda(
            ["typ", "x"],
            Code(
                "$s1 if isinstance($x, $t1) else $s2",
                s1=subcodes[0][1].code,
                s2=subcodes[1][1].code,
                t1=subcodes[0][0],
            ),
        )

    @ovld
    @code_generator
    def f(typ: type[Dataclass], x: Dataclass):
        if x is not All:
            return recurse.resolve(typ, All)
        (typ,) = typ.__args__
        parts = []
        for field in fields(typ):
            lbda = resolve_subcode(
                eval(field.type) if isinstance(field.type, str) else field.type,
                recurse,
            )
            subcode = lbda(Code("UNUSED"), Code(f"$x.{field.name}"))
            parts.append(subcode)
        code = Code("$cons($[, ]parts)", cons=typ, parts=parts, recurse=recurse)
        return Lambda(["typ", "x"], code)

    assert f(int, 3) == 4
    assert f(Point, Point(7, 8)) == Point(8, 9)
    assert f(Tree, Tree(1, 2)) == Tree(2, 3)
    assert f(TwoPoints, TwoPoints(Point(1, 2), Point(7, 8))) == TwoPoints(
        Point(2, 3), Point(8, 9)
    )
    assert f(Empty, Empty()) == Empty()


def test_scoped_subcodes():
    c = Code([Code("a = $x", x=1234), Code("b = $x + $y", x=4321)], y=666)
    assert c.fill() == "a = 1234\nb = 4321 + 666\n"


def test_lambda():
    df = Lambda(Code("$x + $body"), body=1234)
    assert df.create_body(["x", "y"]).fill() == "return x + 1234"
    assert df.create_expression(["x", "y"]).fill() == "x + 1234"


def test_def():
    df = Def(Code("return $x + $body"), body=1234)
    assert df.create_body(["x", "y"]).fill() == "return x + 1234"
    with pytest.raises(ValueError):
        df.create_expression(["x", "y"])


def test_codegen_priority():
    @ovld
    @code_generator
    def f(x: int):
        return Lambda(..., "'A'")

    @ovld
    @code_generator(priority=1)
    def f(x: object):
        return Lambda(..., "'B'")

    assert f(2) == "B"



from numbers import Number
from typing import Literal

import pytest

from ovld.core import OvldBase, ovld
from ovld.dependent import (
    Dependent,
    EndsWith,
    Equals,
    HasKey,
    ParametrizedDependentType,
    Regexp,
    StartsWith,
    dependent_check,
)
from ovld.types import HasMethod, Union
from ovld.utils import UsageError


class Bounded(ParametrizedDependentType):
    def default_bound(self, *parameters):
        return type(parameters[0])

    def check(self, value):
        min, max = self.parameters
        return min <= value <= max

    def __lt__(self, other):
        smin, smax = self.parameters
        omin, omax = other.parameters
        return (omin < smin and omax >= smax) or (omin <= smin and omax > smax)


def test_equality():
    assert Equals[0] == Equals[0]
    assert Equals[0] != Equals[1]
    assert Bounded[0, 10] == Bounded[0, 10]


def test_dependent_type():
    @ovld
    def f(x: Equals[0]):
        return "zero"

    @f.register
    def f(x: Equals[1]):
        return "one"

    @f.register
    def f(x: int):
        return "nah"

    assert f(0) == "zero"
    assert f(1) == "one"
    assert f(2) == "nah"


def test_dependent_func():
    @ovld
    def f(x: Dependent[int, lambda x: x >= 0]):
        return "positive"

    @f.register
    def f(x: Dependent[int, lambda x: x < 0]):
        return "negative"

    assert f(0) == "positive"
    assert f(1) == "positive"
    assert f(2) == "positive"
    assert f(-2) == "negative"


def test_dependent_func():
    @dependent_check
    def InBetween(value: Number, min, max):
        return min <= value <= max

    @ovld
    def f(x: InBetween[1, 10]):
        return "A"

    @ovld
    def f(x: Number):
        return "B"

    @f.register
    def f(x):
        return "C"

    assert f(0) == "B"
    assert f(5) == "A"
    assert f("x") == "C"


def test_dependent_method():
    class Candy(OvldBase):
        def __init__(self, n):
            self.n = n

        @ovld
        def f(self, x: Equals[0]):
            return "zero"

        def f(self, x: Equals[1]):
            return "one"

        def f(self, x: int):
            return "nah"

    c = Candy(3)

    assert c.f(0) == "zero"
    assert c.f(1) == "one"
    assert c.f(2) == "nah"


def test_dependent_ambiguity():
    @ovld
    def f(s: Dependent[str, StartsWith["hell"]]):
        return "A"

    @f.register
    def f(s: Dependent[str, EndsWith["ello"]]):
        return "B"

    assert f("hell") == "A"
    with pytest.raises(TypeError, match="Ambiguous resolution"):
        f("hello")


def test_regexp_isinstance():
    assert isinstance("hello", Regexp[r"[ehl]+"])


def test_regexp():
    @ovld
    def f(s: Regexp[r"hell+o"]):
        return "hello"

    @ovld
    def f(s: str):
        return "bye"

    assert f("hello") == "hello"
    assert f("helllllllllo") == "hello"
    assert f("helllllllll") == "bye"
    assert f("oh helllllo there") == "hello"


def test_has_key():
    @ovld
    def f(d: Dependent[dict, HasKey["a"]]):
        return "a"

    @f.register
    def f(d: Dependent[dict, HasKey["b", "c"]]):
        return "b|c"

    @f.register
    def f(d: dict):
        return "other"

    assert f({"a": 1}) == "a"
    assert f({"b": 2}) == "other"
    assert f({"b": 2, "c": 8}) == "b|c"
    with pytest.raises(TypeError):
        f({"a": 9, "b": 2, "c": 8})


def test_has_key_isinstance():
    assert isinstance({"a": 3}, HasKey["a"])
    assert not isinstance({"b": 3}, HasKey["a"])
    assert isinstance({"a": 7, "b": 3}, HasKey["a", "b"])
    assert not isinstance({"b": 3}, HasKey["a", "b"])


def test_dependent_lists():
    HasLen = HasMethod["__len__"]

    @dependent_check
    def Nonempty(x: HasLen):
        return len(x) > 0

    @dependent_check
    def Length(x: HasLen, n):
        return len(x) == n

    @dependent_check
    def MinLength(x: HasLen, n):
        return len(x) >= n

    @ovld
    def f(li: Nonempty):
        return "nonempty"

    @f.register(priority=1)
    def f(li: Length[3]):
        return "three"

    @f.register(priority=1)
    def f(li: MinLength[5]):
        return ">=five"

    @f.register
    def f(li: HasLen):
        return "other"

    assert f([]) == "other"
    assert f([1]) == "nonempty"
    assert f([1, 2, 3]) == "three"
    assert f([1, 2, 3, 4]) == "nonempty"
    assert f([1, 2, 3, 4, 5]) == ">=five"
    assert f([1, 2, 3, 4, 5, 6, 7, 8]) == ">=five"


def test_bounded():
    @ovld
    def f(x: Dependent[Number, Bounded(0, 10)]):
        return "0-10"

    @f.register
    def f(x: Dependent[Number, Bounded(2, 6)]):
        return "2-6"

    assert Bounded(0, 10) > Bounded(2, 6)
    assert f(1) == "0-10"
    assert f(5) == "2-6"


def test_bound():
    @ovld
    def f(x: Dependent[int, lambda x: 1 / 0]):
        return 123

    @ovld
    def f(x):
        return x

    o = object()
    assert f(o) is o


def test_no_type_bound():
    class XXX(ParametrizedDependentType):
        pass

    with pytest.raises(UsageError):

        @ovld
        def f(x: XXX[1, 2, 3]):
            return "123"

        f(123)


def test_vs_catchall():
    @ovld
    def f(x: Dependent[Number, Bounded[0, 10]]):
        return "0-10"

    @f.register
    def f(x):
        return "other"

    assert f(1) == "0-10"
    assert f(25) == "other"
    assert f("zazz") == "other"


def test_or():
    o = Union[Equals[0], Equals[1]]
    assert isinstance(0, o)
    assert isinstance(1, o)
    assert not isinstance(2, o)

    @ovld
    def f(x: o):
        return 2

    assert f(0) == 2
    assert f(1) == 2


def test_or_mix():
    o = Union[str, Equals[0]]

    assert isinstance("x", o)
    assert isinstance(0, o)
    assert not isinstance(1, o)

    @ovld
    def f(x: o):
        return x

    assert f(0) == 0
    assert f("x") == "x"
    with pytest.raises(TypeError):
        f(1)


def test_and():
    a = Bounded[0, 100] & Bounded[-50, 50]
    assert not isinstance(-50, a)
    assert isinstance(1, a)
    assert not isinstance(100, a)

    @ovld
    def f(x: a):
        return "yes"

    @ovld
    def f(x: object):
        return "no"

    assert f(-50) == "no"
    assert f(1) == "yes"
    assert f(100) == "no"


def test_and_with_static_type():
    @ovld
    def f(x: str & Dependent[object, lambda obj: bool(obj)]):  # type: ignore
        return "yes"

    @ovld
    def f(x: object):
        return "no"

    assert f("hello") == "yes"
    assert f("") == "no"
    assert f([1, 2, 3]) == "no"
    assert f(100) == "no"


def test_keyed_plus_other():
    # Tests a failure mode of generating dispatch code like HANDLER = HANDLERS[x]
    # if you also have to check a condition on y.

    @ovld
    def f(x: Literal[0], y: Bounded[0, 100]):
        return "yes"

    @ovld
    def f(x: Literal[1], y: Bounded[0, 100]):
        return "yes"

    @ovld
    def f(x: Literal[2], y: Bounded[0, 100]):
        return "yes"

    @ovld
    def f(x: Literal[3], y: Bounded[0, 100]):
        return "yes"

    @ovld
    def f(x: Literal[4], y: Bounded[0, 100]):
        return "yes"

    @ovld
    def f(x: int, y: int):
        return "no"

    assert f(0, 50) == "yes"
    assert f(3, 50) == "yes"
    assert f(0, 150) == "no"


class Lettered(type):
    def __instancecheck__(cls, x):
        return cls.letter in (getattr(x, "__name__", None) or str(x))


LetterF = Lettered("LetterF", (), {"letter": "f"})
LetterX = Lettered("LetterX", (), {"letter": "x"})


def test_arbitrary_instancecheck():
    @ovld
    def f(x: LetterX):
        return "X-TREME"

    @ovld
    def f(x: LetterF):
        return "fudged"

    @ovld
    def f(x: object):
        return False

    assert f(next) == "X-TREME"
    assert f("wow") is False
    assert f("extreme") == "X-TREME"
    assert f(filter) == "fudged"



import inspect
import re
from collections.abc import Callable as _Callable
from collections.abc import Mapping, Sequence
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    TypeVar,
)

from .codegen import Code
from .types import (
    Intersection,
    Order,
    clsstring,
    normalize_type,
    subclasscheck,
    typeorder,
)


class DependentType(type):
    __dependent__ = True

    exclusive_type = False
    keyable_type = False
    bound_is_name = False

    def __new__(cls, *args, **kwargs):
        value = super().__new__(cls, cls.__name__, (), {})
        return value

    def __init__(self, bound):
        self.bound = bound

    def __class_getitem__(cls, item):
        items = (item,) if not isinstance(item, tuple) else item
        return cls(*items)

    def with_bound(self, new_bound):  # pragma: no cover
        return type(self)(new_bound)

    def check(self, value):  # pragma: no cover
        raise NotImplementedError()

    def codegen(self):
        return Code("$this.check($arg)", this=self)

    def __type_order__(self, other):
        if isinstance(other, DependentType):
            order = typeorder(self.bound, other.bound)
            if order is Order.SAME:
                # It isn't fully deterministic which of these will be called
                # because of set ordering between the types we compare
                if self < other:  # pragma: no cover
                    return Order.LESS
                elif other < self:  # pragma: no cover
                    return Order.MORE
                else:
                    return Order.NONE
            else:  # pragma: no cover
                return order
        elif subclasscheck(other, self.bound) or subclasscheck(self.bound, other):
            return Order.LESS
        else:
            return Order.NONE

    def __is_supertype__(self, other):
        if isinstance(other, DependentType):
            return False
        elif subclasscheck(other, self.bound):
            return True
        else:
            return False

    def __instancecheck__(self, other):
        return isinstance(other, self.bound) and self.check(other)

    def __lt__(self, other):
        return False

    def __and__(self, other):
        return Intersection[self, other]

    def __rand__(self, other):
        return Intersection[other, self]

    __repr__ = __str__ = clsstring


class ParametrizedDependentType(DependentType):
    def __init__(self, *parameters, bound=None):
        super().__init__(self.default_bound(*parameters) if bound is None else bound)
        self.__args__ = self.parameters = parameters
        self.__origin__ = None
        self.__post_init__()

    def __post_init__(self):
        pass

    @property
    def parameter(self):
        return self.parameters[0]

    def default_bound(self, *parameters):
        return None

    def with_bound(self, new_bound):
        return type(self)(*self.parameters, bound=new_bound)

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.parameters == other.parameters
            and self.bound == other.bound
        )

    def __hash__(self):
        return hash(self.parameters) ^ hash(self.bound)

    def __str__(self):
        if self.bound_is_name:
            origin = self.bound
            bound = ""
        else:
            origin = self
            if self.bound != self.default_bound(*self.parameters):
                bound = f" < {clsstring(self.bound)}"
            else:
                bound = ""
        args = ", ".join(map(clsstring, self.__args__))
        return f"{origin.__name__}[{args}]{bound}"

    __repr__ = __str__


class FuncDependentType(ParametrizedDependentType):
    def default_bound(self, *_):
        return self._default_bound

    def __lt__(self, other):
        if len(self.parameters) != len(other.parameters):
            return False
        p1g = sum(
            p1 is Any and p2 is not Any for p1, p2 in zip(self.parameters, other.parameters)
        )
        p2g = sum(
            p2 is Any and p1 is not Any for p1, p2 in zip(self.parameters, other.parameters)
        )
        return p2g and not p1g

    def check(self, value):
        return type(self).func(value, *self.parameters)


def dependent_check(fn=None, bound_is_name=False):
    if fn is None:
        return partial(dependent_check, bound_is_name=bound_is_name)

    if isinstance(fn, type):
        params = inspect.signature(fn.check).parameters
        bound = normalize_type(
            list(inspect.signature(fn.check).parameters.values())[1].annotation,
            fn.check,
        )
        t = type(
            fn.__name__,
            (FuncDependentType,),
            {
                "bound_is_name": bound_is_name,
                "_default_bound": bound,
                **vars(fn),
            },
        )

    else:
        params = inspect.signature(fn).parameters
        bound = normalize_type(
            list(inspect.signature(fn).parameters.values())[0].annotation, fn
        )
        t = type(
            fn.__name__,
            (FuncDependentType,),
            {
                "func": fn,
                "bound_is_name": bound_is_name,
                "_default_bound": bound,
            },
        )
        if len(params) == 1:
            t = t()

    return t


class Equals(ParametrizedDependentType):
    keyable_type = True

    def default_bound(self, *parameters):
        return type(parameters[0])

    def check(self, value):
        return value in self.parameters

    @classmethod
    def keygen(cls):
        return Code("$arg")

    def get_keys(self):
        return [self.parameter]

    def codegen(self):
        if len(self.parameters) == 1:
            return Code("($arg == $p)", p=self.parameter)
        else:
            return Code("($arg in $ps)", ps=self.parameters)


class ProductType(ParametrizedDependentType):
    bound_is_name = True

    def default_bound(self, *subtypes):
        return tuple

    def check(self, value):
        return (
            isinstance(value, tuple)
            and len(value) == len(self.parameters)
            and all(isinstance(x, t) for x, t in zip(value, self.parameters))
        )

    def codegen(self):
        checks = ["len($arg) == $n"]
        params = {"n": len(self.parameters)}
        for i, p in enumerate(self.parameters):
            checks.append(f"isinstance($arg[{i}], $p{i})")
            params[f"p{i}"] = p
        return Code(" and ".join(checks), params)

    def __type_order__(self, other):
        if isinstance(other, ProductType):
            if len(other.parameters) == len(self.parameters):
                return Order.merge(
                    typeorder(a, b) for a, b in zip(self.parameters, other.parameters)
                )
            else:
                return Order.NONE
        else:
            return NotImplemented


@dependent_check(bound_is_name=True)
def SequenceFastCheck(value: Sequence, typ):
    return not value or isinstance(value[0], typ)


@dependent_check(bound_is_name=True)
def CollectionFastCheck(value: Collection, typ):
    for x in value:
        return isinstance(x, typ)
    else:
        return True


@dependent_check(bound_is_name=True)
def MappingFastCheck(value: Mapping, kt, vt):
    if not value:
        return True
    for k in value:
        break
    return isinstance(k, kt) and isinstance(value[k], vt)


@dependent_check
def Callable(fn: _Callable, argt, rett):
    from .core import Signature

    sig = Signature.extract(fn)
    return (
        sig.max_pos >= len(argt) >= sig.req_pos
        and not sig.req_names
        and all(subclasscheck(t1, t2) for t1, t2 in zip(argt, sig.types))
        and subclasscheck(sig.return_type, rett)
    )


@dependent_check
class HasKey:
    def check(self, value: Mapping):
        return all(k in value for k in self.parameters)

    def codegen(self):
        return Code(
            "($[ and ]checks)", checks=[Code("($k in $arg)", k=k) for k in self.parameters]
        )


@dependent_check
def StartsWith(value: str, prefix):
    return value.startswith(prefix)


@dependent_check
def EndsWith(value: str, suffix):
    return value.endswith(suffix)


@dependent_check
class Regexp:
    def __post_init__(self):
        self.rx = re.compile(self.parameter)

    def check(self, value: str):
        return bool(self.rx.search(value))

    def codegen(self):
        return Code("bool($rx.search($arg))", rx=self.rx)


class Dependent:
    def __class_getitem__(cls, item):
        bound, dt = item
        if not isinstance(dt, DependentType):
            dt = dependent_check(dt)
        return dt.with_bound(bound)


if TYPE_CHECKING:  # pragma: no cover
    from typing import Annotated, TypeAlias

    T = TypeVar("T")
    A = TypeVar("A")
    Dependent: TypeAlias = Annotated[T, A]



import pytest

from ovld import MultiTypeMap
from ovld.core import Signature


class Animal:
    pass


class Mammal(Animal):
    pass


class Bird(Animal):
    pass


class Cat(Mammal):
    pass


class Robin(Bird):
    pass


nada = frozenset()


def mksig(
    types,
    req_pos,
    max_pos,
    req_names=nada,
    vararg=False,
    priority=0,
):
    return Signature(
        types=types,
        return_type=None,
        req_pos=req_pos,
        max_pos=max_pos,
        req_names=req_names,
        vararg=vararg,
        priority=priority,
    )


def _get(tm, *key):
    try:
        return [tm[key]]
    except KeyError as err:
        return [x.handler for x in err.args[1]]


def test_empty():
    tm = MultiTypeMap()

    with pytest.raises(KeyError):
        tm[(object, object)]


def test_not_found():
    tm = MultiTypeMap()

    tm[(int,)] = "X"

    with pytest.raises(KeyError):
        tm[(object,)]


def test_inheritance():
    tm = MultiTypeMap()

    tm.register(mksig((object,), 1, 1), "o")
    tm.register(mksig((Animal,), 1, 1), "A")
    tm.register(mksig((Bird,), 1, 1), "B")

    assert _get(tm, object) == ["o"]
    assert _get(tm, int) == ["o"]
    assert _get(tm, str) == ["o"]

    assert _get(tm, Bird) == ["B"]
    assert _get(tm, Robin) == ["B"]

    assert _get(tm, Animal) == ["A"]
    assert _get(tm, Mammal) == ["A"]
    assert _get(tm, Cat) == ["A"]


def test_multiple_dispatch():
    tm = MultiTypeMap()

    tm.register(mksig((object, object), 2, 2), "oo")
    tm.register(mksig((Animal, object), 2, 2), "Ao")
    tm.register(mksig((Mammal, object), 2, 2), "Mo")
    tm.register(mksig((object, Animal), 2, 2), "oA")
    tm.register(mksig((object, Mammal), 2, 2), "oM")
    tm.register(mksig((Mammal, Mammal), 2, 2), "MM")

    assert _get(tm, int, int) == ["oo"]
    assert _get(tm, Cat, int) == ["Mo"]
    assert _get(tm, int, Cat) == ["oM"]
    assert _get(tm, Cat, Cat) == ["MM"]
    assert _get(tm, Cat, Cat) == ["MM"]
    assert _get(tm, Robin, int) == ["Ao"]
    assert _get(tm, Cat, Robin) == ["Mo", "oA"]
    assert set(_get(tm, Robin, Robin)) == {"Ao", "oA"}


def test_direct_ambiguity():
    tm = MultiTypeMap()

    tm.register(mksig((str, int), 2, 2), "A")
    tm.register(mksig((str, int), 2, 2), "B")

    # Could be in either order
    assert set(_get(tm, str, int)) == {"A", "B"}


def test_priority_same_signature():
    tm = MultiTypeMap()

    tm.register(mksig((str, int), 2, 2, nada, False, 1), "A")
    tm.register(mksig((str, int), 2, 2), "B")

    assert _get(tm, str, int) == ["A"]


def test_priority():
    tm = MultiTypeMap()

    tm.register(mksig((object,), 1, 1), "o")
    tm.register(mksig((Animal,), 1, 1), "A")
    tm.register(mksig((Mammal,), 1, 1), "M")
    tm.register(mksig((Cat,), 1, 1), "C")
    tm.register(mksig((Bird,), 1, 1, nada, False, 1), "B")  # <= higher priority
    tm.register(mksig((Robin,), 1, 1), "R")

    assert _get(tm, object) == ["o"]
    assert _get(tm, int) == ["o"]
    assert _get(tm, str) == ["o"]

    assert _get(tm, Bird) == ["B"]
    assert _get(tm, Robin) == ["B"]  # <= Bird's higher priority matters here

    assert _get(tm, Animal) == ["A"]
    assert _get(tm, Mammal) == ["M"]
    assert _get(tm, Cat) == ["C"]


def test_caching():
    tm = MultiTypeMap()

    tm.register(mksig((Mammal, Mammal), 2, 2), "X")

    assert (Mammal, Mammal) not in tm
    assert (Cat, Cat) not in tm

    assert _get(tm, Cat, Cat) == ["X"]
    # Second one should be cached and should return the same thing
    assert _get(tm, Cat, Cat) == ["X"]

    assert (Cat, Cat) in tm

    assert _get(tm, Mammal, Mammal) == ["X"]
    # Second one should be cached and should return the same thing
    assert _get(tm, Mammal, Mammal) == ["X"]

    assert (Mammal, Mammal) in tm


def test_deeper_caching():
    tm = MultiTypeMap()

    sig = mksig((Mammal, Mammal), 2, 2)
    tm.register(sig, "MM")

    m0 = tm.maps[0]
    m1 = tm.maps[1]

    assert Cat not in m0
    assert Cat not in m1

    assert _get(tm, Cat, Cat) == ["MM"]

    assert m0[Cat] == {("MM", sig): 0}
    assert m1[Cat] == {("MM", sig): 0}


def test_cache_invalidation():
    tm = MultiTypeMap()

    tm.register(mksig((Mammal, Mammal), 2, 2), "MM")
    assert (Cat, Cat) not in tm
    assert _get(tm, Cat, Cat) == ["MM"]

    assert (Cat, Cat) in tm
    tm.register(mksig((Cat, Cat), 2, 2), "CC")
    assert (Cat, Cat) not in tm
    assert _get(tm, Cat, Cat) == ["CC"]



"""Miscellaneous utilities."""

import builtins
import functools
import re
import typing
from abc import ABCMeta
from itertools import count

_builtins_dict = vars(builtins)


try:
    from types import UnionType

    UnionTypes = (type(typing.Union[int, str]), UnionType)

except ImportError:  # pragma: no cover
    UnionType = None
    UnionTypes = (type(typing.Union[int, str]),)


class Named:
    """A named object.

    This class can be used to construct objects with a name that will be used
    for the string representation.
    """

    def __init__(self, name):
        """Construct a named object.

        Arguments:
            name: The name of this object.
        """
        self.name = name

    def __repr__(self):
        """Return the object's name."""
        return self.name


BOOTSTRAP = Named("BOOTSTRAP")
MISSING = Named("MISSING")


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=None, **kwargs):
        if fn is None:

            @functools.wraps(deco)
            def newer_deco(fn):
                return deco(fn, **kwargs)

            return newer_deco
        else:
            return deco(fn, **kwargs)

    return new_deco


class CodegenInProgress(Exception):
    pass


class UsageError(Exception):
    pass


class ResolutionError(TypeError):
    pass


class SpecialForm:
    def __init__(self, name, message=None):
        self.__name = name
        self.__message = (
            message or f"{name}() can only be used from inside an @ovld-registered function."
        )

    def __call__(self, *args, **kwargs):
        raise UsageError(self.__message)

    def __getattr__(self, attr):
        raise UsageError(self.__message)

    def __str__(self):
        return f"<SpecialForm {self.__name}>"

    __repr__ = __str__


class GenericAliasMC(type):
    def __instancecheck__(cls, obj):
        return hasattr(obj, "__origin__")


class GenericAlias(metaclass=GenericAliasMC):
    pass


def clsstring(cls):
    if cls is object:
        return "*"
    elif args := typing.get_args(cls):
        origin = typing.get_origin(cls) or cls
        args = ", ".join(map(clsstring, args))
        return f"{origin.__name__}[{args}]"
    else:
        r = repr(cls)
        if r.startswith("<class ") or r.startswith("<enum "):
            return cls.__name__
        else:
            return r


def typemap_entry_string(cls):
    if isinstance(cls, tuple):
        key, typ = cls
        return f"{key}: {clsstring(typ)}"
    else:
        return clsstring(cls)


def sigstring(types):
    return ", ".join(map(typemap_entry_string, types))


def subtler_type(obj):
    if isinstance(obj, GenericAlias):
        return type[obj]
    elif isinstance(obj, UnionTypes):
        return type[obj]
    elif obj is typing.Any:
        return type[object]
    elif isinstance(obj, type):
        return type[obj]
    else:
        return type(obj)


class NameDatabase:
    def __init__(self, default_name="TMP"):
        self.default_name = default_name
        self.count = count()
        self.variables = {}
        self.names = {}
        self.registered = set()

    def register(self, name):
        self.registered.add(name)

    def gensym(self, desired_name, value=None):
        i = 1
        name = desired_name
        while name in self.registered or (
            name in _builtins_dict and _builtins_dict[name] != value
        ):
            name = f"{desired_name}{i}"
            i += 1
        self.registered.add(name)
        return name

    def get(self, value, suggested_name=None):
        if type(value) in (int, float, str):
            return repr(value)
        if id(value) in self.names:
            return self.names[id(value)]
        dflt = suggested_name or self.default_name
        if (
            isinstance(value, GenericAlias) and typing.get_origin(value) is type
        ):  # pragma: no cover
            name = "t_" + getattr(typing.get_args(value)[0], "__name__", dflt)
        else:
            name = getattr(value, "__name__", dflt)
        if not re.match(string=name, pattern=r"^[a-zA-Z_][a-zA-Z0-9_]*$"):
            name = dflt
        name = self.gensym(name, value)
        self.variables[name] = value
        self.names[id(value)] = name
        return name

    __getitem__ = get


def get_args(tp):
    args = getattr(tp, "__args__", None)
    if not isinstance(args, tuple):
        args = ()
    return args


_standard_instancechecks = {
    type.__instancecheck__,
    GenericAlias.__instancecheck__,
    type(list[object]).__instancecheck__,
    ABCMeta.__instancecheck__,
    type(typing.Protocol).__instancecheck__,
    type(typing.Any).__instancecheck__,
}


def is_dependent(t):
    if any(is_dependent(subt) for subt in get_args(t)):
        return typing.get_origin(t) is not type
    elif hasattr(t, "__dependent__"):
        return t.__dependent__
    elif not isinstance(t, type):
        return False
    elif type(t).__instancecheck__ not in _standard_instancechecks:
        return True
    return False



import importlib
import inspect
import sys
import typing
from dataclasses import dataclass
from functools import partial
from typing import Protocol, get_args, runtime_checkable

from .codegen import Code
from .mro import Order, TypeRelationship, subclasscheck, typeorder
from .recode import generate_checking_code
from .typemap import TypeMap
from .utils import UnionType, UnionTypes, UsageError, clsstring, get_args

NoneType = type(None)


def eval_annotation(t, ctx, locals, catch=False):
    try:
        if isinstance(t, str):
            if hasattr(ctx, "__globals__"):
                glb = ctx.__globals__
            elif hasattr(ctx, "__module__"):  # pragma: no cover
                glb = vars(importlib.import_module(ctx.__module__))
            else:  # pragma: no cover
                glb = {}
            return eval(t, glb, locals)
        else:  # pragma: no cover
            return t
    except Exception:  # pragma: no cover
        if catch:
            return None
        raise


class TypeNormalizer:
    def __init__(self, generic_handlers=None):
        self.generic_handlers = generic_handlers or TypeMap()

    def register_generic(self, generic, handler=None):
        if handler is None:
            return partial(self.register_generic, generic)
        else:
            self.generic_handlers.register(generic, handler)

    def __call__(self, t, fn, lcl={}):
        from .dependent import DependentType

        if isinstance(t, str):
            t = eval_annotation(t, fn, lcl)

        if t is type:
            t = type[object]
        elif t is typing.Any:
            t = object
        elif t is inspect._empty:
            t = object
        elif t in UnionTypes:
            return type[t]
        elif isinstance(t, typing._AnnotatedAlias):
            t = t.__origin__

        origin = getattr(t, "__origin__", None)
        if UnionType and isinstance(t, UnionType):
            return self(t.__args__, fn)
        elif origin is type:
            return t
        elif origin and getattr(t, "__args__", None) is None:
            return t
        elif origin is not None:
            try:
                results = self.generic_handlers[origin]
                results = list(results.items())
                results.sort(key=lambda x: x[1], reverse=True)
                assert results
                rval = results[0][0](self, t, fn)
                if isinstance(origin, type) and hasattr(rval, "with_bound"):
                    rval = rval.with_bound(origin)
                return rval
            except KeyError:  # pragma: no cover
                raise TypeError(f"ovld does not understand generic type {t}")
        elif isinstance(t, tuple):
            return Union[tuple(self(t2, fn) for t2 in t)]
        elif isinstance(t, DependentType) and not t.bound:
            raise UsageError(
                f"Dependent type {t} has not been given a type bound. Please use Dependent[<bound>, {t}] instead."
            )
        elif t is None:
            return NoneType
        else:
            return t


normalize_type = TypeNormalizer()


@normalize_type.register_generic(typing.Union)
def _(self, t, fn):
    return self(t.__args__, fn)


class MetaMC(type):
    __dependent__ = False

    def __new__(T, name, handler):
        return super().__new__(T, name, (), {"_handler": handler})

    def __init__(cls, name, handler):
        cls.__args__ = get_args(handler)

    def codegen(cls):
        return cls._handler.codegen()

    def __type_order__(cls, other):
        return cls._handler.__type_order__(other)

    def __is_supertype__(cls, other):
        return cls._handler.__is_supertype__(other)

    def __is_subtype__(cls, other):  # pragma: no cover
        return cls._handler.__is_subtype__(other)

    def __subclasscheck__(cls, sub):
        return cls._handler.__subclasscheck__(sub)

    def __instancecheck__(cls, obj):
        return cls._handler.__instancecheck__(obj)

    def __eq__(cls, other):
        return (
            type(cls) is type(other)
            and type(cls._handler) is type(other._handler)
            and cls._handler == other._handler
        )

    def __hash__(cls):
        return hash(cls._handler)

    def __and__(cls, other):
        return Intersection[cls, other]

    def __rand__(cls, other):
        return Intersection[other, cls]

    def __str__(self):
        return str(self._handler)

    __repr__ = __str__


class SingleFunctionHandler:
    def __init__(self, handler, args):
        self.handler = handler
        self.args = self.__args__ = args

    def __type_order__(self, other):
        results = self.handler(other, *self.args)
        if isinstance(results, TypeRelationship):
            return results.order
        else:
            return NotImplemented

    def __is_supertype__(self, other):
        results = self.handler(other, *self.args)
        if isinstance(results, bool):
            return results
        elif isinstance(results, TypeRelationship):
            return results.supertype
        else:  # pragma: no cover
            return NotImplemented

    def __is_subtype__(self, other):  # pragma: no cover
        results = self.handler(other, *self.args)
        if isinstance(results, TypeRelationship):
            return results.subtype
        else:
            return NotImplemented

    def __subclasscheck__(self, sub):
        return self.__is_supertype__(sub)

    def __instancecheck__(self, obj):
        return issubclass(type(obj), self)

    def __str__(self):
        args = ", ".join(map(clsstring, self.__args__))
        return f"{self.handler.__name__}[{args}]"


def class_check(condition):
    """Return a class with a subclassing relation defined by condition.

    For example, a dataclass is a subclass of `class_check(dataclasses.is_dataclass)`,
    and a class which name starts with "X" is a subclass of
    `class_check(lambda cls: cls.__name__.startswith("X"))`.

    Arguments:
        condition: A function that takes a class as an argument and returns
            True or False depending on whether it matches some condition.
    """
    return MetaMC(condition.__name__, SingleFunctionHandler(condition, ()))


def parametrized_class_check(fn):
    """Return a parametrized class checker.

    In essence, parametrized_class_check(fn)[X] will call fn(cls, X) in order
    to check whether cls matches the condition defined by fn and X.

    Arguments:
        fn: A function that takes a class and one or more additional arguments,
            and returns True or False depending on whether the class matches.
    """

    class _C:
        def __class_getitem__(_, arg):
            if not isinstance(arg, tuple):
                arg = (arg,)

            if isinstance(fn, type):
                return MetaMC(fn.__name__, fn(*arg))
            else:
                return MetaMC(fn.__name__, SingleFunctionHandler(fn, arg))

    _C.__name__ = fn.__name__
    _C.__qualname__ = fn.__qualname__
    return _C


def _getcls(ref):
    module, *parts = ref.split(".")
    curr = __import__(module)
    for part in parts:
        curr = getattr(curr, part)
    return curr


class AllMC(type):
    __bound__ = False

    def __type_order__(self, other):
        return Order.MORE

    def __is_subtype__(self, other):
        if not self.__bound__:
            return True
        return subclasscheck(self.__bound__, other) or subclasscheck(other, self.__bound__)

    def __is_supertype__(self, other):
        return False

    def __subclasscheck__(self, other):  # pragma: no cover
        return False

    def __isinstance__(self, other):  # pragma: no cover
        return False

    def __getitem__(self, item):
        name = getattr(item, "__name__", str(item))
        return type(f"All[{name}]", (self,), {"__bound__": item})


class All(metaclass=AllMC):
    """All is the empty/void/bottom type -- it acts as a subtype of all types.

    It is basically the opposite of Any: nothing is an instance of All. The main
    place you want to use All is as a wildcard in contravariant settings, e.g.
    all 2-argument functions are instances of Callable[[All, All], Any] because
    the arguments are contravariant.
    """


class WhateverMC(AllMC):
    def __is_supertype__(self, other):
        return True

    def __subclasscheck__(self, other):  # pragma: no cover
        return True

    def __isinstance__(self, other):  # pragma: no cover
        return True


class Whatever(metaclass=WhateverMC):
    """This type is a superclass and a subclass of everything.

    It's not a coherent type, more like a convenience.

    It'll match anything anywhere, so you can write e.g.
    Callable[[Whatever, Whatever], Whatever] to match any function of
    two arguments. Any only works in covariant settings, and All only
    works in contravariant settings.
    """


class Deferred:
    """Represent a class from an external module without importing it.

    For instance, `Deferred["numpy.ndarray"]` matches instances of
    numpy.ndarray, but it does not import numpy. When tested against a
    class, if the first part of class's `__module__` is `numpy`, then
    we do get the class and perform a normal issubclass check.

    If the module is already loaded, `Deferred` returns the class directly.

    Arguments:
        ref: A string starting with a module name representing the path
            to import a class.
    """

    def __class_getitem__(cls, ref):
        module, _ = ref.split(".", 1)
        if module in sys.modules:
            return _getcls(ref)

        def check(cls):
            full_cls_mod = getattr(cls, "__module__", None)
            cls_module = full_cls_mod.split(".", 1)[0] if full_cls_mod else None
            if cls_module == module:
                return issubclass(cls, _getcls(ref))
            else:
                return False

        return MetaMC(f"Deferred[{ref}]", SingleFunctionHandler(check, ()))


@parametrized_class_check
def Exactly(cls, base_cls):
    """Match the class but not its subclasses."""
    return TypeRelationship(
        order=Order.LESS if cls is base_cls else typeorder(base_cls, cls),
        supertype=cls is base_cls,
    )


@parametrized_class_check
def StrictSubclass(cls, base_cls):
    """Match subclasses but not the base class."""
    return isinstance(cls, type) and issubclass(cls, base_cls) and cls is not base_cls


@parametrized_class_check
class Union:
    def __init__(self, *types):
        self.__args__ = self.types = types

    def codegen(self):
        return Code("($[ or ]checks)", checks=[generate_checking_code(t) for t in self.types])

    def __type_order__(self, other):
        if other is Union:
            return Order.LESS
        classes = self.types
        compare = [x for t in classes if (x := typeorder(t, other)) is not Order.NONE]
        if not compare:
            return Order.NONE
        elif any(x is Order.MORE or x is Order.SAME for x in compare):
            return Order.MORE
        else:
            return Order.LESS

    def __is_supertype__(self, other):
        return any(subclasscheck(other, t) for t in self.types)

    def __is_subtype__(self, other):
        if other is Union:
            return True
        return NotImplemented  # pragma: no cover

    def __subclasscheck__(self, sub):
        return self.__is_supertype__(sub)

    def __instancecheck__(self, obj):
        return any(isinstance(obj, t) for t in self.types)

    def __eq__(self, other):
        return self.__args__ == other.__args__

    def __hash__(self):
        return hash(self.__args__)

    def __str__(self):
        return " | ".join(map(clsstring, self.__args__))


@parametrized_class_check
class Intersection:
    def __init__(self, *types):
        self.__args__ = self.types = types

    def codegen(self):
        return Code("($[ and ]checks)", checks=[generate_checking_code(t) for t in self.types])

    def __type_order__(self, other):
        if other is Intersection:
            return Order.LESS
        classes = self.types
        compare = [x for t in classes if (x := typeorder(t, other)) is not Order.NONE]
        if not compare:
            return Order.NONE
        elif any(x is Order.LESS or x is Order.SAME for x in compare):
            return Order.LESS
        else:
            return Order.MORE

    def __is_supertype__(self, other):
        return all(subclasscheck(other, t) for t in self.types)

    def __is_subtype__(self, other):  # pragma: no cover
        if other is Intersection:
            return True
        return NotImplemented

    def __subclasscheck__(self, sub):
        return self.__is_supertype__(sub)

    def __instancecheck__(self, obj):
        return all(isinstance(obj, t) for t in self.types)

    def __eq__(self, other):
        return self.__args__ == other.__args__

    def __hash__(self):
        return hash(self.__args__)

    def __str__(self):
        return " & ".join(map(clsstring, self.__args__))


@parametrized_class_check
def HasMethod(cls, method_name):
    """Match classes that have a specific method."""
    return hasattr(cls, method_name)


@runtime_checkable
@dataclass
class Dataclass(Protocol):
    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "__dataclass_fields__") and hasattr(
            subclass, "__dataclass_params__"
        )



import pytest

from .common import function_builder, ovld_dispatch, with_functions

###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_tweaknum(dispatch):
    @dispatch
    def tweaknum(n: int, *, add: int):
        return n + add

    @dispatch
    def tweaknum(n: int, *, mul: int):
        return n * mul

    @dispatch
    def tweaknum(n: int, *, pow: int):
        return n**pow

    return tweaknum


#######
# ifs #
#######


def tweaknum_ifs(n, *, add=None, mul=None, pow=None):
    assert isinstance(n, int)
    assert (add is not None) + (mul is not None) + (pow is not None) == 1
    if add is not None:
        assert isinstance(add, int)
        return n + add
    if mul is not None:
        assert isinstance(mul, int)
        return n * mul
    if pow is not None:
        assert isinstance(pow, int)
        return n**pow


###################
# match statement #
###################


def tweaknum_match(n, **kwargs):
    match kwargs:
        case {"add": int(x)}:
            return n + x
        case {"mul": int(x)}:
            return n * x
        case {"pow": int(x)}:
            return n**x


####################
# Test definitions #
####################


@pytest.mark.benchmark(group="tweaknum")
@with_functions(
    ovld=make_tweaknum(ovld_dispatch),
    custom__ifs=tweaknum_ifs,
    custom__match=tweaknum_match,
)
def test_tweaknum(fn, benchmark):
    def run():
        return fn(10, add=3), fn(5, mul=7), fn(2, pow=5)

    result = benchmark(run)
    assert result == (13, 35, 32)



import pytest

from .common import (
    function_builder,
    multimethod_dispatch,
    multipledispatch_dispatch,
    ovld_dispatch,
    plum_dispatch,
    runtype_dispatch,
    with_functions,
)

###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_add(dispatch):
    @dispatch
    def add(x: list, y: list):
        return [add(a, b) for a, b in zip(x, y)]

    @dispatch
    def add(x: tuple, y: tuple):
        return tuple(add(a, b) for a, b in zip(x, y))

    @dispatch
    def add(x: dict, y: dict):
        return {k: add(v, y[k]) for k, v in x.items()}

    @dispatch
    def add(x: object, y: object):
        return x + y

    return add


##############
# isinstance #
##############


def add_isinstance(x, y):
    if isinstance(x, dict) and isinstance(y, dict):
        return {k: add_isinstance(v, y[k]) for k, v in x.items()}
    elif isinstance(x, tuple) and isinstance(y, tuple):
        return tuple(add_isinstance(a, b) for a, b in zip(x, y))
    elif isinstance(x, list) and isinstance(y, list):
        return [add_isinstance(a, b) for a, b in zip(x, y)]
    else:
        return x + y


###################
# match statement #
###################


def add_match(x, y):
    match (x, y):
        case ({}, {}):
            return {k: add_match(v, y[k]) for k, v in x.items()}
        case (tuple(), tuple()):
            return tuple(add_match(a, b) for a, b in zip(x, y))
        case ([*x], [*y]):
            return [add_match(a, b) for a, b in zip(x, y)]
        case _:
            return x + y


####################
# Test definitions #
####################

A = {"xs": list(range(50)), "ys": ("o", (6, 7))}
B = {"xs": list(range(10, 60)), "ys": ("x", (7, 6))}
C = {"xs": list(range(10, 110, 2)), "ys": ("ox", (13, 13))}


@pytest.mark.benchmark(group="add")
@with_functions(
    ovld=make_add(ovld_dispatch),
    plum=make_add(plum_dispatch),
    multimethod=make_add(multimethod_dispatch),
    multipledispatch=make_add(multipledispatch_dispatch),
    runtype=make_add(runtype_dispatch),
    custom__isinstance=add_isinstance,
    custom__match=add_match,
)
def test_add(fn, benchmark):
    result = benchmark(fn, A, B)
    assert result == C



import pytest

from ovld import call_next, recurse
from ovld.utils import UsageError


def test_unusable_special_form():
    with pytest.raises(UsageError, match="recurse.. can only be used from inside an @ovld"):
        recurse()

    with pytest.raises(UsageError, match="call_next.. can only be used from inside an @ovld"):
        call_next()

    with pytest.raises(UsageError, match="recurse.. can only be used from inside an @ovld"):
        recurse.next


def test_special_form_str():
    assert str(recurse) == repr(recurse) == "<SpecialForm recurse>"



import ast
import inspect
import sys
from types import NoneType

import pytest

from .common import (
    function_builder,
    multimethod_dispatch,
    multipledispatch_dispatch,
    ovld_dispatch,
    plum_dispatch,
    runtype_dispatch,
    singledispatch_dispatch,
    with_functions,
)


def foo(xs, ys):
    zs = [x + y for x, y in zip(xs, ys)]
    zs.append("the beginning")
    return zs


def bar(xs, ys):
    zs = [x**y for x, y in zip(xs, ys)]
    zs.append("the end")
    return zs


@pytest.fixture(scope="module")
def foobar():
    return ast.parse(inspect.getsource(foo)), ast.parse(inspect.getsource(bar))


@function_builder
def make_transform(dispatch):
    @dispatch
    def transform(node: list):
        return [transform(x) for x in node]

    @dispatch
    def transform(node: int | str | NoneType):
        return node

    @dispatch
    def transform(node: ast.AST):
        kw = {field: transform(getattr(node, field)) for field in node._fields}
        return type(node)(**kw)

    @dispatch
    def transform(node: ast.BinOp):
        return ast.BinOp(
            op=ast.Pow(),
            left=transform(node.left),
            right=transform(node.right),
        )

    @dispatch
    def transform(node: ast.Constant):
        return ast.Constant(
            value=node.value.replace("beginning", "end")
            if isinstance(node.value, str)
            else node.value,
            kind=node.kind,
        )

    return transform


###################
# NodeTransformer #
###################


class NT(ast.NodeTransformer):
    def visit_BinOp(self, node):
        return ast.BinOp(
            op=ast.Pow(),
            left=self.visit(node.left),
            right=self.visit(node.right),
        )

    def visit_Constant(self, node):
        return ast.Constant(
            value=node.value.replace("beginning", "end")
            if isinstance(node.value, str)
            else node.value,
            kind=node.kind,
        )


####################
# Test definitions #
####################


@pytest.mark.benchmark(group="ast")
@with_functions(
    ovld=make_transform(ovld_dispatch),
    plum=make_transform(plum_dispatch),
    multimethod=make_transform(multimethod_dispatch),
    multipledispatch=make_transform(multipledispatch_dispatch),
    runtype=make_transform(runtype_dispatch),
    singledispatch=make_transform(singledispatch_dispatch),
    custom=NT().visit,
)
def test_ast(fn, foobar, benchmark):
    if (
        sys.version_info < (3, 11)
        and getattr(fn, "_dispatch", None) is singledispatch_dispatch
    ):
        pytest.skip()
    result = benchmark(fn, foobar[0])
    assert ast.dump(result, indent=2) == ast.dump(foobar[1], indent=2).replace("bar", "foo")



from numbers import Number
from typing import Literal

import pytest

from .common import (
    function_builder,
    multimethod_dispatch,
    ovld_dispatch,
    plum_dispatch,
    with_functions,
)

###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_calc(dispatch):
    @dispatch
    def calc(num: Number):
        return num

    @dispatch
    def calc(tup: tuple):
        return calc(*tup)

    @dispatch
    def calc(op: Literal["add"], x: object, y: object):
        return calc(x) + calc(y)

    @dispatch
    def calc(op: Literal["sub"], x: object, y: object):
        return calc(x) - calc(y)

    @dispatch
    def calc(op: Literal["mul"], x: object, y: object):
        return calc(x) * calc(y)

    @dispatch
    def calc(op: Literal["div"], x: object, y: object):
        return calc(x) / calc(y)

    @dispatch
    def calc(op: Literal["pow"], x: object, y: object):
        return calc(x) ** calc(y)

    @dispatch
    def calc(op: Literal["sqrt"], x: object):
        return calc(x) ** 0.5

    return calc


#########
# match #
#########


def calc_match(expr):
    match expr:
        case ("add", x, y):
            return calc_match(x) + calc_match(y)
        case ("sub", x, y):
            return calc_match(x) - calc_match(y)
        case ("mul", x, y):
            return calc_match(x) * calc_match(y)
        case ("div", x, y):
            return calc_match(x) / calc_match(y)
        case ("pow", x, y):
            return calc_match(x) ** calc_match(y)
        case ("sqrt", x):
            return calc_match(x) ** 0.5
        case Number():
            return expr


#########
# dict #
#########


def _add(x, y):
    return calc_dict(x) + calc_dict(y)


def _sub(x, y):
    return calc_dict(x) - calc_dict(y)


def _mul(x, y):
    return calc_dict(x) * calc_dict(y)


def _div(x, y):
    return calc_dict(x) / calc_dict(y)


def _pow(x, y):
    return calc_dict(x) ** calc_dict(y)


def _sqrt(x):
    return calc_dict(x) ** 0.5


ops = {
    "add": _add,
    "sub": _sub,
    "mul": _mul,
    "div": _div,
    "pow": _pow,
    "sqrt": _sqrt,
}


def calc_dict(expr):
    if isinstance(expr, tuple):
        return ops[expr[0]](*expr[1:])
    elif isinstance(expr, Number):
        return expr


####################
# Test definitions #
####################

expr = ("add", ("mul", ("sqrt", 4), 7), ("div", ("add", 6, 4), ("sub", 5, 3)))
expected_result = 19


@pytest.mark.benchmark(group="calc")
@with_functions(
    ovld=make_calc(ovld_dispatch),
    plum=make_calc(plum_dispatch),
    multimethod=make_calc(multimethod_dispatch),
    custom__dict=calc_dict,
    custom__match=calc_match,
)
def test_calc(fn, benchmark):
    result = benchmark(fn, expr)
    assert result == expected_result



import inspect
import linecache
import re
import textwrap
from ast import _splitlines_no_ff as splitlines
from itertools import count
from types import FunctionType

from .utils import MISSING, NameDatabase, keyword_decorator, sigstring

_current = count()


def rename_code(co, newname):  # pragma: no cover
    if hasattr(co, "replace"):
        if hasattr(co, "co_qualname"):
            return co.replace(co_name=newname, co_qualname=newname)
        else:
            return co.replace(co_name=newname)
    else:
        return type(co)(
            co.co_argcount,
            co.co_kwonlyargcount,
            co.co_nlocals,
            co.co_stacksize,
            co.co_flags,
            co.co_code,
            co.co_consts,
            co.co_names,
            co.co_varnames,
            co.co_filename,
            newname,
            co.co_firstlineno,
            co.co_lnotab,
            co.co_freevars,
            co.co_cellvars,
        )


def transfer_function(
    func,
    argdefs=MISSING,
    closure=MISSING,
    code=MISSING,
    globals=MISSING,
    name=MISSING,
):
    closure = func.__closure__ if closure is MISSING else closure
    if closure:
        new_fn = FunctionType(
            argdefs=func.__defaults__ if argdefs is MISSING else argdefs,
            closure=func.__closure__ if closure is MISSING else closure,
            code=func.__code__ if code is MISSING else code,
            globals=func.__globals__ if globals is MISSING else globals,
            name=func.__name__ if name is MISSING else name,
        )
    else:
        new_fn = FunctionType(
            argdefs=func.__defaults__ if argdefs is MISSING else argdefs,
            code=func.__code__ if code is MISSING else code,
            globals=func.__globals__ if globals is MISSING else globals,
            name=func.__name__ if name is MISSING else name,
        )
    new_fn.__kwdefaults__ = func.__kwdefaults__
    new_fn.__annotations__ = func.__annotations__
    new_fn.__dict__.update(func.__dict__)
    return new_fn


def rename_function(fn, newname):
    """Create a copy of the function with a different name."""
    return transfer_function(
        func=fn,
        code=rename_code(fn.__code__, newname),
        name=newname,
    )


def instantiate_code(symbol, code, inject={}):
    virtual_file = f"<ovld:{abs(hash(code)):x}>"
    linecache.cache[virtual_file] = (None, None, splitlines(code), virtual_file)
    code = compile(source=code, filename=virtual_file, mode="exec")
    glb = {**inject}
    exec(code, glb, glb)
    return glb[symbol]


# # Previous version: generate a temporary file
# def instantiate_code(symbol, code, inject={}):
#     tf = tempfile.NamedTemporaryFile("w")
#     _tempfiles.append(tf)
#     tf.write(code)
#     tf.flush()
#     glb = runpy.run_path(tf.name)
#     rval = glb[symbol]
#     rval.__globals__.update(inject)
#     return rval


subr = re.compile(r"\$(|=|:)(|\[[^\[\]]+\])([a-zA-Z0-9_]+)")
symr = re.compile(r"[a-zA-Z0-9_]")


def sub(template, subs):
    idx = next(_current)

    def repl_fn(m):
        prefix, sep, name = m.groups()
        value = subs(name, bool(sep))
        if value is None:
            return m.group()
        if sep:
            value = sep[1:-1].join(value)
        if prefix == "=" and not symr.fullmatch(value):
            return f"{name}__{idx}"
        if prefix == ":" and not symr.fullmatch(value):
            return f"({name}__{idx} := {value})"
        return value

    return subr.sub(string=template, repl=repl_fn)


def format_code(code, indent=0, nl=False):
    if isinstance(code, str):
        return f"{code}\n" if nl and not code.endswith("\n") else code
    elif isinstance(code, (list, tuple)):
        lines = [format_code(line, indent + 4, True) for line in code]
        block = "".join(lines)
        return textwrap.indent(block, " " * indent)
    else:  # pragma: no cover
        raise TypeError(f"Cannot format code from type {type(code)}")


def _gensym():
    return f"__G{next(_current)}"


class Code:
    def __init__(self, template, substitutions={}, **substitutions_kw):
        self.template = template
        self.substitutions = {**substitutions, **substitutions_kw}

    def sub(self, **subs):
        return Code(self.template, self.substitutions, **subs)

    def defaults(self, subs):
        return Code(self.template, subs, **self.substitutions)

    def _mapsub(self, template, code_recurse, getsub):
        if isinstance(template, (list, tuple)):
            return [self._mapsub(t, code_recurse, getsub) for t in template]

        if isinstance(template, Code):
            return code_recurse(template)

        return sub(template, getsub)

    def rename(self, subs):
        def getsub(name, sep=False):
            return f"${renaming[name]}" if name in renaming else None

        subs = {k: v for k, v in subs.items() if k not in self.substitutions}

        renaming = {k: _gensym() for k in subs}
        new_subs = {renaming[k]: v for k, v in subs.items()}

        def _rename_step(v):
            if isinstance(v, Code):
                return v.rename(subs)
            elif isinstance(v, (list, tuple)):
                if any(isinstance(x, Code) for x in v):
                    return [_rename_step(x) for x in v]
                else:
                    return v
            else:
                return v

        new_subs.update({k: _rename_step(v) for k, v in self.substitutions.items()})
        new_template = self._mapsub(self.template, lambda c: c.rename(subs), getsub)
        return Code(new_template, new_subs)

    def fill(self, ndb=None):
        if ndb is None:
            ndb = NameDatabase()

        def make(name, v, sep):
            if isinstance(v, Code):
                return v.defaults(self.substitutions).fill(ndb)
            elif sep:
                return [make(name, x, sep) for x in v]
            else:
                return ndb.get(v, suggested_name=name)

        def getsub(name, sep=False):
            if name in self.substitutions:
                return make(name, self.substitutions[name], sep)
            else:  # pragma: no cover
                return None

        result = self._mapsub(
            self.template, lambda c: c.defaults(self.substitutions).fill(ndb), getsub
        )
        return format_code(result)


class Function:
    def __init__(self, args, code=None, **subs):
        if code is None:
            code = args
            args = ...
        self.args = args
        if subs:
            self.code = code.sub(**subs) if isinstance(code, Code) else Code(code, subs)
        else:
            self.code = code if isinstance(code, Code) else Code(code)

    def rename(self, *args):
        return self.code.rename(dict(zip(self.args, args)))

    def __call__(self, *args):
        self.args = args if self.args is ... else self.args
        args = [Code(name) if isinstance(name, str) else name for name in args]
        return self.rename(*args)


class Def(Function):
    def create_body(self, argnames):
        return self(*argnames)

    def create_expression(self, argnames):
        raise ValueError("Cannot convert Def to expression")


class Lambda(Function):
    def create_body(self, argnames):
        return Code("return $body", body=self(*argnames))

    def create_expression(self, argnames):
        return self(*argnames)


def regen_signature(fn, ndb):  # pragma: no cover
    sig = inspect.signature(fn)
    args = []
    ko_flag = False
    po_flag = False
    for argname, arg in sig.parameters.items():
        if argname == "cls":
            argname = "self"
        ndb.register(argname)
        argstring = argname
        if arg.default is not inspect._empty:
            argstring += f" = {ndb[arg.default]}"

        if arg.kind is inspect._KEYWORD_ONLY:
            if not ko_flag:
                ko_flag = True
                args.append("*")
        elif arg.kind is inspect._VAR_POSITIONAL:
            ko_flag = True
        elif arg.kind is inspect._VAR_KEYWORD:
            raise TypeError("**kwargs are not accepted")
        elif arg.kind is inspect._POSITIONAL_OR_KEYWORD:
            if po_flag:
                args.append("/")
                po_flag = False
        elif arg.kind is inspect._POSITIONAL_ONLY:
            po_flag = True
        args.append(argstring)

    if po_flag:
        args.append("/")

    return ", ".join(args)


fgen_template = """
def {fn}({args}):
{body}
"""


def codegen_specializer(typemap, fn, tup):
    is_method = typemap.ovld.argument_analysis.is_method
    ndb = NameDatabase(default_name="INJECT")
    args = regen_signature(fn, ndb)
    body = fn(typemap.ovld.specialization_self, *tup) if is_method else fn(*tup)
    cg = None
    if isinstance(body, Function):
        cg = body
        argnames = [
            "self" if arg == "cls" else arg for arg in inspect.signature(fn).parameters
        ]
        body = body.create_body(argnames)
    if isinstance(body, Code):
        body = body.fill(ndb)
    elif isinstance(body, str):
        pass
    elif isinstance(body, FunctionType):
        return body
    elif body is None:
        return None
    body = textwrap.indent(body, "    ")
    code = fgen_template.format(fn="__GENERATED__", args=args, body=body)
    func = instantiate_code("__GENERATED__", code, inject=ndb.variables)
    adjusted_name = f"{fn.__name__.split('[')[0]}[{sigstring(tup)}]"
    func = rename_function(func, adjusted_name)
    func.__codegen__ = cg
    func.__orig_name__ = fn.__name__
    return func


@keyword_decorator
def code_generator(fn, priority=0):
    fn.specializer = codegen_specializer
    if priority:
        fn.priority = priority
    return fn



import sys
from functools import singledispatch

import pytest
from multimethod import multimethod as multimethod_dispatch
from multipledispatch import dispatch as _md_dispatch
from plum import dispatch as plum_dispatch
from runtype import multidispatch as runtype_dispatch

from ovld import ovld as ovld_dispatch


def _locate(fn):
    fr = sys._getframe(1)
    while fr and fn.__code__ not in fr.f_code.co_consts:
        fr = fr.f_back
    return fr.f_locals.get(fn.__name__, None)


def _getanns(fn):
    anns = fn.__annotations__.values()
    anns = [tuple(ann.__args__) if hasattr(ann, "__args__") else ann for ann in anns]
    return anns


def multipledispatch_dispatch(fn):
    anns = _getanns(fn)
    existing = _locate(fn)
    if existing:
        existing.register(*anns)(fn)
        return existing
    else:
        return _md_dispatch(*anns)(fn)


def singledispatch_dispatch(fn):
    existing = _locate(fn)
    if existing:
        existing.register(fn)
        return existing
    else:
        return singledispatch(fn)


def with_functions(**tests):
    return pytest.mark.parametrize(
        "fn",
        argvalues=list(tests.values()),
        ids=list(tests.keys()),
    )


def function_builder(defn):
    def wrapper(dispatch):
        try:
            rval = defn(dispatch)
        except Exception as exc:
            e = exc

            def reraise(*args, **kwargs):
                raise e

            rval = reraise
        try:
            rval._dispatch = dispatch
        except AttributeError:
            pass
        return rval

    return wrapper


__all__ = [
    "multimethod_dispatch",
    "plum_dispatch",
    "runtype_dispatch",
    "ovld_dispatch",
    "multipledispatch_dispatch",
    "singledispatch_dispatch",
    "function_builder",
    "with_functions",
]



import re

import pytest

from ovld.dependent import Regexp

from .common import (
    function_builder,
    ovld_dispatch,
    plum_dispatch,
    with_functions,
)

###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_regexp(dispatch):
    @dispatch
    def regexp(x: Regexp[r"^a"]):
        return "one"

    @dispatch
    def regexp(x: Regexp[r"^[bcd]"]):
        return "two"

    @dispatch
    def regexp(x: Regexp[r"end$"]):
        return "three"

    return regexp


##############
# isinstance #
##############


def regexp_search(x):
    if re.search(string=x, pattern=r"^a"):
        return "one"
    if re.search(string=x, pattern=r"^[bcd]"):
        return "two"
    if re.search(string=x, pattern=r"end$"):
        return "three"


re1 = re.compile(r"^a")
re2 = re.compile(r"^[bcd]")
re3 = re.compile(r"end$")


def regexp_compiled(x):
    if re1.search(x):
        return "one"
    if re2.search(x):
        return "two"
    if re3.search(x):
        return "three"


def regexp_compiled_nonexclusive(x):
    r1 = bool(re1.search(x))
    r2 = bool(re2.search(x))
    r3 = bool(re3.search(x))
    if r1:
        return "one"
    if r2:
        return "two"
    if r3:
        return "three"


####################
# Test definitions #
####################


@pytest.mark.benchmark(group="regexp")
@with_functions(
    ovld=make_regexp(ovld_dispatch),
    plum=make_regexp(plum_dispatch),
    custom__search=regexp_search,
    custom__compiled=regexp_compiled,
    custom__compiled_nx=regexp_compiled_nonexclusive,
)
def test_regexp(fn, benchmark):
    def run():
        return [
            fn("allo"),
            fn("canada"),
            fn("the end"),
        ]

    result = benchmark(run)
    assert result == ["one", "two", "three"]



import os
import sys
from dataclasses import dataclass

from ovld import ovld
from ovld.types import (
    Dataclass,
    Deferred,
    Exactly,
    HasMethod,
    Intersection,
    Order,
    StrictSubclass,
    Union,
    class_check,
    parametrized_class_check,
    typeorder,
)


@dataclass
class Point:
    x: int
    y: int


def test_meta():
    class Apple:
        pass

    class Banana:
        pass

    class Brownie:
        pass

    class Butterscotch:
        pass

    class Cherry:
        pass

    @class_check
    def B_(cls):
        return cls.__name__.startswith("B")

    @parametrized_class_check
    def ClassPrefix(cls, prefix):
        return cls.__name__.startswith(prefix)

    @ovld
    def f(x):
        return "no B"

    @f.register
    def f(x: Brownie):
        return "Brownie"

    @f.register
    def f(x: B_):
        return "B!"

    @f.register
    def f(x: Butterscotch):
        return "Butterscotch"

    @f.register
    def f(x: ClassPrefix["A"]):
        return "Almost B"

    assert f(Apple()) == "Almost B"
    assert f(Banana()) == "B!"
    assert f(Brownie()) == "Brownie"
    assert f(Butterscotch()) == "Butterscotch"
    assert f(Cherry()) == "no B"


def test_deferred_builtins():
    assert Deferred["builtins.object"] is object
    assert Deferred["builtins.TypeError"] is TypeError


def test_deferred():
    assert "gingerbread" not in sys.modules
    sys.path.append(os.path.join(os.path.dirname(__file__), "modules"))

    @ovld
    def f(x: Deferred["gingerbread.House"]):
        return "Gingerbread house!"

    @f.register
    def f(x):
        return "object"

    assert "gingerbread" not in sys.modules

    import gingerbread

    assert f(gingerbread.House()) == "Gingerbread house!"

    assert "gingerbread" in sys.modules


def test_exactly():
    class Fruit:
        pass

    class Apple(Fruit):
        pass

    assert str(Exactly[Fruit]) == "Exactly[Fruit]"

    @ovld
    def f(x: Exactly[Fruit]):
        return "yes"

    @f.register
    def f(x: object):
        return "no"

    assert f(Fruit()) == "yes"
    assert f(Apple()) == "no"


def test_strict_subclass():
    class Fruit:
        pass

    class Apple(Fruit):
        pass

    @ovld
    def f(x: StrictSubclass[Fruit]):
        return "yes"

    @f.register
    def f(x: object):
        return "no"

    assert f(Fruit()) == "no"
    assert f(Apple()) == "yes"


def test_Dataclass():
    @ovld
    def f(x: Dataclass):
        return "yes"

    @f.register
    def f(x: object):
        return "no"

    @f.register
    def f(x: type[Dataclass]):
        return "type, yes"

    @f.register
    def f(x: type[object]):
        return "type, no"

    assert f(Point(1, 2)) == "yes"
    assert f(1234) == "no"
    assert f(Point) == "type, yes"
    assert f(int) == "type, no"


def test_intersection():
    class A:
        pass

    class B:
        pass

    class C(A, B):
        pass

    @ovld
    def f(x: A):
        return "A"

    @ovld
    def f(x: B):
        return "B"

    @ovld
    def f(x: Union[A, B]):
        return "A | B"

    @ovld
    def f(x: Intersection[A, B]):
        return "A & B"

    @ovld
    def f(x):
        return "other"

    assert f(A()) == "A"
    assert f(B()) == "B"
    assert f(C()) == "A & B"
    assert f(object()) == "other"
    assert f(1.5) == "other"


def test_hasmethod():
    assert isinstance([1, 2, 3], HasMethod["__len__"])


def test_intersection_operator():
    class Base:
        def f(self):
            pass

    class X(Base):
        def g(self):
            pass

    class Y(Base):
        def h(self):
            pass

    @ovld
    def f(x: HasMethod["f"] & HasMethod["g"]):  # type: ignore
        return "yes"

    @ovld
    def f(x: Base & HasMethod["h"]):  # type: ignore
        return "yeah"

    @ovld
    def f(x: object):
        return "no"

    assert f(X()) == "yes"
    assert f(Y()) == "yeah"


def test_intersection_issubclass():
    assert not issubclass(int, Intersection[int, str])


def test_intersection_isinstance():
    assert not isinstance(1234, Intersection[int, str])
    assert isinstance(1234, Intersection[int])


def test_intersection_typeorder():
    assert typeorder(Intersection[int, str], str) is Order.LESS
    assert typeorder(str, Intersection[int, str]) is Order.MORE
    assert typeorder(float, Intersection[int, str]) is Order.NONE
    assert typeorder(Intersection[float], Intersection[int, str]) is Order.NONE
    assert typeorder(Intersection[str], Intersection[int, str]) is Order.MORE
    assert typeorder(Intersection, Intersection[int, str]) is Order.MORE


def test_union_issubclass():
    assert issubclass(int, Union[int, str])


def test_union_isinstance():
    assert isinstance(1234, Union[int, str])
    assert not isinstance(1234, Union[float, str])


def test_union_typeorder():
    assert typeorder(Union[int, str], str) is Order.MORE
    assert typeorder(str, Union[int, str]) is Order.LESS
    assert typeorder(float, Union[int, str]) is Order.NONE
    assert typeorder(Union[float], Union[int, str]) is Order.NONE
    assert typeorder(Union[str], Union[int, str]) is Order.LESS
    assert typeorder(Union, Union[int, str]) is Order.MORE



import inspect
import re
import sys
import typing
from dataclasses import dataclass
from typing import Annotated, Any

import pytest

from ovld import (
    Dataclass,
    Ovld,
    OvldBase,
    OvldMC,
    call_next,
    current_code,
    extend_super,
    is_ovld,
    ovld,
    recurse,
    resolve,
)
from ovld.dependent import Dependent, Equals, Regexp, StartsWith
from ovld.types import UnionTypes
from ovld.utils import MISSING, ResolutionError, UsageError

from .test_typemap import Animal, Bird, Mammal


def test_Ovld():
    o = Ovld()

    @o.register
    def f(x: int):
        """Integers!"""
        return "int"

    @o.register  # noqa: F811
    def f(x: float):
        """Floats!"""
        return "float"

    assert is_ovld(o)

    assert f(2) == "int"
    assert f(2.0) == "float"

    with pytest.raises(TypeError):
        f(object())

    with pytest.raises(TypeError):
        f()


def test_nargs():
    @ovld
    def f():
        return 0

    @ovld
    def f(x: int):
        return 1

    @ovld
    def f(x: int, y: int):
        return 2

    @ovld
    def f(x: int, y: int, z: int):
        return 3

    assert f() == 0
    assert f(0) == 1
    assert f(0, 0) == 2
    assert f(0, 0, 0) == 3


def test_multimethod():
    o = Ovld()

    @o.register
    def f(x: object, y: object):
        return "oo"

    @o.register
    def f(x: int, y):
        return "io"

    @o.register
    def f(x, y: int):
        return "oi"

    @o.register
    def f(x: str, y: str):
        return "ss"

    assert f(1.0, 1.0) == "oo"

    assert f(1, object()) == "io"
    assert f(1, "x") == "io"

    assert f("x", 1) == "oi"
    assert f(object(), 1) == "oi"

    assert f("a", "b") == "ss"

    with pytest.raises(TypeError):
        # Ambiguous
        f(1, 2)

    @o.register
    def f(x: int, y: int):
        return "ii"

    assert f(1, 2) == "ii"


def test_redefine():
    o = Ovld()

    @o.register
    def f(x: int):
        return x + x

    assert f(10) == 20

    @o.register
    def f(x: int):
        return x * x

    assert f(10) == 100


def test_redefine_2():
    o = Ovld()

    @o.register
    def f(x: Animal):
        return "animal"

    assert f(Bird()) == "animal"
    assert f(Mammal()) == "animal"

    @o.register
    def f(x: Mammal):
        return "mammal"

    assert f(Bird()) == "animal"
    assert f(Mammal()) == "mammal"


def test_redefine_parent():
    o = Ovld()
    o2 = o.copy()

    @o.register
    def f(x: Animal):
        return "animal"

    @o2.register
    def f2(x: Bird):
        return "a bird"

    assert f(Bird()) == "animal"
    assert f(Mammal()) == "animal"

    @o.register
    def f(x: Mammal):
        return "mammal"

    assert f2(Bird()) == "a bird"
    assert f2(Mammal()) == "mammal"

    with pytest.raises(Exception):

        @o.register
        def f(x: Bird):
            return "b bird"


def test_redefine_parent_with_linkback():
    o = Ovld()
    o2 = o.copy(linkback=True)

    @o.register
    def f(x: Animal):
        return "animal"

    @o2.register
    def f2(x: Bird):
        return "a bird"

    assert f(Bird()) == "animal"
    assert f(Mammal()) == "animal"

    @o.register
    def f(x: Mammal):
        return "mammal"

    assert f2(Bird()) == "a bird"
    assert f2(Mammal()) == "mammal"

    @o.register
    def f(x: Bird):
        return "b bird"

    assert f(Bird()) == "b bird"
    assert f2(Bird()) == "a bird"


def test_typetuple():
    o = Ovld()

    @o.register
    def f(x, y):
        return "oo"

    @o.register
    def f(x: (int, float), y):
        return "io"

    @o.register
    def f(x, y: (int, float)):
        return "oi"

    @o.register
    def f(x: (int, float), y: (int, float)):
        return "ii"

    assert f(1, 1) == "ii"
    assert f(1.0, 1.0) == "ii"

    assert f(1, "x") == "io"
    assert f("x", 1) == "oi"


def test_typetuple_override():
    o = Ovld()

    @o.register
    def f(x: (int, float)):
        return "if"

    @o.register
    def f(x: int):
        return "i"

    assert f(1) == "i"
    assert f(1.0) == "if"


def test_union():
    o = Ovld()

    @o.register
    def f(x, y):
        return "oo"

    @o.register
    def f(x: typing.Union[int, float], y):
        return "io"

    @o.register
    def f(x, y: typing.Union[int, float]):
        return "oi"

    @o.register
    def f(x: typing.Union[int, float], y: typing.Union[int, float]):
        return "ii"

    assert f(1, 1) == "ii"
    assert f(1.0, 1.0) == "ii"

    assert f(1, "x") == "io"
    assert f("x", 1) == "oi"


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="union syntax requires python3.10 or higher",
)
def test_union_pipe_syntax():
    o = Ovld()

    @o.register
    def f(x, y):
        return "oo"

    @o.register
    def f(x: int | float, y):
        return "io"

    @o.register
    def f(x, y: int | float):
        return "oi"

    @o.register
    def f(x: int | float, y: int | float):
        return "ii"

    assert f(1, 1) == "ii"
    assert f(1.0, 1.0) == "ii"

    assert f(1, "x") == "io"
    assert f("x", 1) == "oi"


def test_optional():
    o = Ovld()

    @o.register
    def f(x: typing.Optional[int]):
        return "iN"

    @o.register
    def f(x: float):
        return "f"

    @o.register
    def f(x: object):
        return "o"

    assert f(1) == "iN"
    assert f(None) == "iN"
    assert f(1.0) == "f"


def test_none():
    @ovld
    def f(x: None):
        return "null"

    @ovld
    def f(x: object):
        return "X"

    assert f(None) == "null"
    assert f(123) == "X"


def test_abstract_types():
    from collections.abc import Iterable

    o = Ovld()

    @o.register
    def f(x: object):
        return "o"

    @o.register
    def f(xs: Iterable):
        return type(xs)(f(x) for x in xs)

    assert f(1) == "o"
    assert f([1, 2, 3]) == ["o", "o", "o"]
    assert f((1, [2, 3])) == ("o", ["o", "o"])


@typing.runtime_checkable
class CanFly(typing.Protocol):
    def fly(self): ...


@dataclass
class Oiseau:
    feathers: int

    def fly(self):
        return "fly!!"


def test_protocol():
    o = Ovld()

    @o.register
    def f(x: CanFly):
        return "f"

    @o.register
    def f(x):
        return "o"

    assert f(1) == "o"
    assert f(Oiseau(feathers=13)) == "f"


def test_abstract_ambiguity():
    o = Ovld()

    @o.register
    def f(x: CanFly):
        return "f"

    @o.register
    def f(x: Dataclass):
        return "dc"

    @o.register
    def f(x):
        return "o"

    with pytest.raises(TypeError):
        f(Oiseau(feathers=13))


def test_next_abstract_ambiguity():
    o = Ovld()

    @o.register(priority=10)
    def f(x: object):
        return f.next(x)

    @o.register
    def f(x: CanFly):
        return "f"

    @o.register
    def f(x: Dataclass):
        return "dc"

    @o.register
    def f(x):
        return "o"

    with pytest.raises(TypeError):
        f(Oiseau(feathers=13))


@pytest.mark.xfail(reason="varargs are a little broken currently")
def test_varargs():
    o = Ovld()

    @o.register
    def f(*args):
        return "A"

    @o.register
    def f(x: int, *args):
        return "B"

    @o.register
    def f(x: int, y: int, *args):
        return "C"

    assert f() == "A"
    assert f("X") == "A"

    assert f(1) == "B"
    assert f(1, "x", "y") == "B"

    assert f(1, 2) == "C"
    assert f(1, 2, 3) == "C"


def test_default_args():
    o = Ovld()

    @o.register
    def f(x: int):
        return "A"

    @o.register
    def f(x: str, y: int = 3):
        return y

    @o.register
    def f(x: float, y=7):
        return x + y

    assert f(1) == "A"
    assert f("X") == 3
    assert f("X", 24) == 24
    with pytest.raises(TypeError):
        f("X", "Y")

    assert f(3.7) == 10.7


def test_mixins():
    f = Ovld()

    @f.register
    def f(t: int):
        return t + 1

    g = Ovld()

    @g.register
    def g(t: str):
        return t.upper()

    h = Ovld(mixins=[f, g])

    assert f(15) == 16
    with pytest.raises(TypeError):
        f("hello")

    assert g("hello") == "HELLO"
    with pytest.raises(TypeError):
        g(15)

    assert h(15) == 16
    assert h("hello") == "HELLO"


def test_mixin_through_variant():
    f = Ovld()

    @f.register
    def f(t: int):
        return t + 1

    g = Ovld()

    @g.register
    def g(t: str):
        return t.upper()

    @g.variant(mixins=[f])
    def h(t: object):
        return "OBJ"

    assert h(15) == 16
    assert h("hello") == "HELLO"
    assert h(1.5) == "OBJ"


def test_recurse():
    f = Ovld()

    @f.register
    def f(xs: list):
        return [recurse(x) for x in xs]

    @f.register
    def f(x: int):
        return x + 1

    assert f([1, 2, 3]) == [2, 3, 4]


def test_recurse_method():
    class C(OvldBase):
        def f(self, xs: list):
            return [recurse(x) for x in xs]

        def f(self, x: int):
            return x + 1

    assert C().f([1, 2, 3]) == [2, 3, 4]


def test_recurse_nested():
    @ovld
    def f(xs: list):
        return recurse(sum(recurse(x) for x in xs))

    @ovld
    def f(x: str):
        return len(x)

    @ovld
    def f(x: int):
        return x * x

    assert f(["a", "bbb", "cc"]) == 36


def test_call_next():
    f = Ovld()

    @f.register
    def f(x: int):
        return call_next(x + 1)

    @f.register
    def f(x: object):
        return x * 2

    assert f(3) == 8


def test_call_next_unrelated():
    f = Ovld()

    @f.register
    def f(x: int):
        return f.next(str(x))

    @f.register
    def f(x: str):
        return x * 2

    @f.register
    def f(x: object):
        return "no"

    assert f(3) == "33"
    for k, v in f.map.items():
        print(k, v)


def test_call_next_same_priority():
    f = Ovld()

    @f.register
    def f(x: int):
        return x * 2

    @f.register
    def f(x: int):
        return call_next(x + 1)

    @f.register
    def f(x: int):
        return call_next(-x)

    assert f(5) == -8


def test_recurse_renamed():
    f = Ovld()

    @f.register
    def f(xs: list):
        rec = recurse
        return [rec(x) for x in xs]

    @f.register
    def f(x: int):
        return x + 1

    assert f([1, 2, 3]) == [2, 3, 4]


def test_call_next_must_be_called():
    f = Ovld()

    with pytest.raises(UsageError, match="call_next should be called right away"):

        @f.register
        def f(xs: list):
            cn = call_next
            return [cn(x) for x in xs]

        f.compile()


def test_recurse_closure():
    f = Ovld()

    @f.register
    def f(xs: list):
        def inner():
            return [recurse(x) for x in xs]

        return inner()

    @f.register
    def f(x: int):
        return x + 1

    assert f([1, 2, 3]) == [2, 3, 4]


def test_variant():
    f = Ovld()

    @f.register
    def f(xs: list):
        return [recurse(x) for x in xs]

    @f.register
    def f(x: int):
        return x + 1

    @f.register
    def f(x: object):
        return "A"

    assert f([1, 2, "xxx", [3, 4]]) == [2, 3, "A", [4, 5]]

    @f.variant
    def g(x: object):
        return "B"

    # This does not interfere with f
    assert f([1, 2, "xxx", [3, 4]]) == [2, 3, "A", [4, 5]]

    # The new method in g is used
    assert g([1, 2, "xxx", [3, 4]]) == [2, 3, "B", [4, 5]]


def test_next():
    f = Ovld()

    @f.register
    def f(x: int):
        if x >= 0:
            return x
        else:
            return f.next(x)

    @f.register
    def f(xs: list):
        return [f(x) for x in xs]

    @f.register
    def f(x: object):
        return "OBJECT"

    assert f([-1, 0, 1]) == ["OBJECT", 0, 1]


def test_next_long_chain():
    f = Ovld()

    @f.register
    def f(x: int):
        return ["A", f.next(x)]

    @f.register
    def f(x: object):
        return ["B", f.next(x)]

    @f.register(priority=-1)
    def f(x: object):
        return ["C", f.next(x)]

    @f.register(priority=-2)
    def f(x: object):
        return ["D", x]

    assert f(12) == ["A", ["B", ["C", ["D", 12]]]]


def test_next_bottom():
    f = Ovld()

    @f.register
    def f(x: int):
        if x >= 0:
            return x
        else:
            return f.next(x)

    @f.register
    def f(xs: list):
        return [f(x) for x in xs]

    with pytest.raises(TypeError):
        f([-1, 0, 1])


def test_next_different():
    f = Ovld()

    @f.register
    def f(x: int):
        return f.next(str(x))

    @f.register
    def f(x: str):
        return x * 2

    assert f(5) == "55"


def test_next_nodispatch():
    f = Ovld()

    @f.register
    def f(x: int):
        if x >= 0:
            return x
        else:
            return f.next(x)

    @f.register
    def f(xs: list):
        return [f(x) for x in xs]

    @f.register
    def f(x: object):
        return "OBJECT"

    assert f([-1, 0, 1]) == ["OBJECT", 0, 1]


def test_priority():
    f = Ovld()

    @f.register(priority=1)
    def f(x: object):
        return ["TOP", f.next(x)]

    @f.register
    def f(xs: list):
        return [f(x) for x in xs]

    @f.register
    def f(x: int):
        return x + 1

    @f.register
    def f(x: object):
        return "BOTTOM"

    assert f([1, "x"]) == ["TOP", [["TOP", 2], ["TOP", "BOTTOM"]]]


def test_priority_tuple():
    f = Ovld()

    @f.register(priority=(1, 2))
    def f(x: object):
        return ["TOP", call_next(x)]

    @f.register(priority=(1, 1))
    def f(x: object):
        return "BOTTOM"

    assert f(123) == ["TOP", "BOTTOM"]


def test_resolve_for_values():
    f = Ovld()

    @f.register
    def f(x: int):
        return x * 2

    assert f.resolve_for_values(8)("hello") == "hellohello"


def test_resolve():
    @ovld(priority=1)
    def f(x: int):
        return x * 2

    @ovld(priority=0)
    def f(x: int):
        return x * 3

    f1 = f.resolve(int)
    f2a = f.resolve(int, after=f1)
    f2b = f.resolve(int, after=f1.__code__)
    assert f1("hello") == "hellohello"
    assert f2a("hello") == "hellohellohello"
    assert f2b("hello") == "hellohellohello"


def test_resolve_special_form():
    @ovld(priority=1)
    def f(x: int):
        return resolve(str)(x)

    @ovld(priority=0)
    def f(x: str):
        return x * 3

    assert f(3) == 9


def test_current_code_special_form():
    @ovld(priority=1)
    def f(x: int):
        return resolve(int, after=current_code)(x)

    @ovld(priority=0)
    def f(x: int):
        return x * 3

    assert f(3) == 9


def test_method():
    class Greatifier:
        def __init__(self, n):
            self.n = n

        perform = Ovld()

        @perform.register
        def perform(self, x: int):
            return x + self.n

        @perform.register
        def perform(self, x: str):
            return x + "s" * self.n

    g = Greatifier(2)
    assert g.perform(7) == 9
    assert g.perform("cheese") == "cheesess"


def test_metaclass():
    class Greatifier(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        def perform(self, x: int):
            return x + self.n

        def perform(self, x: str):
            return x + "s" * self.n

        def perform(self, x: object):
            return x

    g = Greatifier(2)
    assert g.perform(7) == 9
    assert g.perform("cheese") == "cheesess"
    assert g.perform(g) is g


def test_metaclass_inherit():
    class Greatifier(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        def perform(self, x: int):
            return x + self.n

    class Greatestifier(Greatifier):
        @extend_super
        def perform(self, x: str):
            return x + "s" * self.n

        def perform(self, x: object):
            return x

    g = Greatifier(2)
    assert g.perform(7) == 9
    with pytest.raises(TypeError):
        g.perform("cheese")

    gg = Greatestifier(2)
    assert gg.perform(7) == 9
    assert gg.perform("cheese") == "cheesess"


def test_metaclass_multiple_inherit():
    class One(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

    class Two(One):
        @extend_super
        def perform(self, x: int):
            return x + self.n

        def perform(self, x: float):
            return x * self.n

    class Three(One):
        @extend_super
        def perform(self, x: str):
            return x + "s" * self.n

    class Four(Two, Three):
        pass

    g = Four(2)
    assert g.perform(7) == 9
    assert g.perform(7.0) == 14
    assert g.perform("cheese") == "cheesess"

    with pytest.raises(TypeError):
        Two(2).perform("cheese")

    with pytest.raises(TypeError):
        Three(2).perform(7)


def test_multiple_inherit_2():
    class One(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        def perform(self, x: int):
            return x + self.n

    class M1:
        @extend_super
        def perform(self, x: float):
            return x * self.n

    class M2:
        @extend_super
        def perform(self, x: str):
            return x + "s" * self.n

    cls = One.create_subclass(M1, M2, name="Test")

    g = cls(2)
    assert g.perform(7) == 9
    assert g.perform(7.0) == 14
    assert g.perform("cheese") == "cheesess"


def test_multiple_inherit_3():
    class One(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        def perform(self, x: int):
            return x + self.n

    class Two:
        @ovld
        def perform(self, x: float):
            return x * self.n

    class Three:
        def perform(self, x: str):
            return x + "s" * self.n

    class Four(One, Two, Three):
        @extend_super
        def perform(self, x: int, y: int):
            return x * y

    g = Four(2)
    assert g.perform(7) == 9
    assert g.perform(7.0) == 14
    assert g.perform("cheese") == "cheesess"
    assert g.perform(5, 4) == 20


def test_metaclass_dispatch():
    class One(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        @ovld(priority=1000)
        def perform(self, x):
            return call_next(x) * 2

        def perform(self, x: int):
            return x + self.n

        def perform(self, xs: list):
            return [self.perform(x) for x in xs]

    x = One(1)
    assert x.perform([1, 2, 3]) == [4, 6, 8, 4, 6, 8]

    class Two(One):
        @extend_super
        def perform(self, x: float):
            return x * self.n

    x2 = Two(2)
    assert x2.perform([1, 2, 3]) == [6, 8, 10, 6, 8, 10]
    assert x2.perform([1.5, 2.5]) == [6.0, 10.0, 6.0, 10.0]


def test_metaclass_dispatch_2():
    class One(metaclass=OvldMC):
        def __init__(self, n):
            self.n = n

        @ovld
        def perform(self, x: int):
            return x + self.n

        def perform(self, xs: list):
            return [self.perform(x) for x in xs]

    x = One(1)
    assert x.perform([1, 2, 3]) == [2, 3, 4]

    class Two(One):
        @extend_super
        @ovld(priority=1000)
        def perform(self, x):
            return call_next(x) * 2

    x2 = Two(1)
    assert x2.perform([1, 2, 3]) == [4, 6, 8, 4, 6, 8]


def test_multiple_dispatch_3():
    class One(OvldBase):
        def __init__(self, n):
            self.n = n

        @ovld(priority=1000)
        def perform(self, x):
            return call_next(x) * 2

        def perform(self, xs: list):
            return [self.perform(x) for x in xs]

    class M1:
        @extend_super
        def perform(self, x: int):
            return x + self.n

    class M2(OvldBase):
        @extend_super
        def perform(self, x: float):
            return x * self.n

        def perform(self, x: str):
            return x + "s" * self.n

    cls = One.create_subclass(M1, M2, name="Test")

    g = cls(2)
    assert g.perform(7) == 18
    assert g.perform(7.0) == 28
    assert g.perform("cheese") == "cheesesscheesess"
    assert g.perform([1, 2.0]) == [6, 8.0, 6, 8.0]


def test_repr():
    humptydumpty = Ovld()

    @humptydumpty.register
    def humptydumpty(x: int):
        return x

    @humptydumpty.register
    def ignore_this_name(x: str):
        return x

    assert humptydumpty.name.endswith(".humptydumpty")
    r = repr(humptydumpty)
    assert r.startswith("<Ovld ")
    assert r.endswith(".humptydumpty>")


def test_repr_misc():
    assert repr(MISSING) == "MISSING"


def test_bizarre_bug():
    # Don't ask

    class A:
        pass

    class B(A, metaclass=OvldMC):
        pass

    assert not isinstance(B(), type)


def test_replacement():
    @ovld
    def f(x: int):
        return 1

    @f.register
    def f2(x: int):
        return 2

    assert f(5) == 2


def test_unregister():
    @ovld
    def f(xs: list):
        return [f(x) for x in xs]

    def intf(x: int):
        return x * x

    f.register(intf)
    assert f([-2, -1, 0, 1, 2, 3]) == [4, 1, 0, 1, 4, 9]
    f.unregister(intf)

    with pytest.raises(TypeError):
        f([-2, -1, 0, 1, 2, 3])


@pytest.mark.xfail(
    condition="platform.python_implementation() == 'PyPy'",
    reason="codefind dependency fails here",
)
def test_conform():
    @ovld
    def f(xs: list):
        return [f(x) for x in xs]

    def intf(x: int):
        return x * x

    def intf2(x: int):
        return x * x if x > 0 else 0

    f.register(intf)
    assert f([-2, -1, 0, 1, 2, 3]) == [4, 1, 0, 1, 4, 9]

    f.map[(int,)]._conformer.__conform__(intf2)
    assert f([-2, -1, 0, 1, 2, 3]) == [0, 0, 0, 1, 4, 9]

    f.map[(int,)]._conformer.__conform__(None)
    with pytest.raises(TypeError):
        f([-2, -1, 0, 1, 2, 3])


@pytest.mark.xfail(
    condition="platform.python_implementation() == 'PyPy'",
    reason="codefind dependency fails here",
)
def test_conform_2():
    @ovld
    def f(xs: list):
        return [f(x) for x in xs]

    def intf(x: int):
        return x * x

    def floatf(x: float):
        return x - 1

    f.register(intf)
    assert f([-2, -1, 0, 1, 2, 3]) == [4, 1, 0, 1, 4, 9]

    with pytest.raises(TypeError):
        f([-2.0, 5.5])

    f.map[(int,)]._conformer.__conform__(floatf)

    with pytest.raises(TypeError):
        f([-2, -1, 0, 1, 2, 3])

    assert f([-2.0, 5.5]) == [-3.0, 4.5]


def test_type_argument():
    class One:
        pass

    class Two(One):
        pass

    class Three(One):
        pass

    @ovld
    def f(t: type[int], x):
        return x * x

    @f.register
    def _f(t: type[str], x):
        return f"hello, {x}"

    @f.register
    def _f(t: type[One], x):
        return 1

    @f.register
    def _f(t: type[Three], x):
        return 3

    assert f(int, 5) == 25
    assert f(str, 5) == "hello, 5"
    assert f(Three, 5) == 3
    assert f(Two, 5) == 1
    assert f(One, 5) == 1


def test_generic_type_argument():
    @ovld
    def f(t: type[list]):
        return "list"

    @f.register
    def f(t: type[dict]):
        return "dict"

    assert f(list) == "list"
    assert f(list[int]) == "list"

    assert f(dict) == "dict"
    assert f(dict[str, int]) == "dict"


def test_annotated_type_argument():
    @ovld
    def f(t: type[Annotated[str, "hello"]]):
        return "hello"

    @ovld
    def f(t: type[Annotated[str, "world"]]):
        return "world"

    @ovld
    def f(t: object):
        return False

    assert f(Annotated[str, "hello"]) == "hello"
    assert f(Annotated[str, "world"]) == "world"
    assert f(Annotated[str, "hello", "X"]) == "hello"
    assert f(Annotated[str, 1, 2, 3, "world"]) == "world"


@pytest.mark.skipif(
    sys.version_info < (3, 12), reason="Python 3.11 is more strict on Annotated"
)
def test_annotated_type_argument_any():
    @ovld
    def f(t: type[Annotated[Any, "hello"]]):
        return "hello"

    @ovld
    def f(t: object):
        return False

    assert f(Annotated[str, "hello"]) == "hello"
    assert f(Annotated[123, "hello", "X"]) == "hello"
    assert f(Annotated[str, 1, 2, 3, "hello"]) == "hello"


@dataclass(frozen=True)
class Anno:
    name: str
    annotation_priority: int = 0


def test_annotated_multiple_ann():
    hello = Anno("hello", 1)
    world = Anno("hello", 2)
    foo = Anno("foo", 1)
    bar = Anno("bar", 1)

    @ovld
    def f(t: type[Annotated[str, hello]]):
        return "hello"

    @ovld
    def f(t: type[Annotated[str, world]]):
        return "world"

    @ovld
    def f(t: type[Annotated[str, foo]]):
        return "foo"

    @ovld
    def f(t: type[Annotated[object, bar]]):
        return "bar"

    @ovld
    def f(t: object):
        return False

    assert f(Annotated[str, hello, world]) == "world"
    assert f(Annotated[str, hello, bar]) == "hello"
    with pytest.raises(ResolutionError):
        assert f(Annotated[str, hello, foo]) == "world"
    assert f("ah") is False


def test_any():
    @ovld
    def f(x: int):
        return "no"

    @ovld
    def f(t: typing.Any):
        return "yes"

    assert f(1) == "no"
    assert f("xyz") == "yes"
    assert f(object()) == "yes"


def test_type_any():
    @ovld
    def f(t: type[dict]):
        return "no"

    @ovld
    def f(t: type[object]):
        return "yes"

    assert f(typing.Any) == "yes"


UT = type(typing.Union[int, str])


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="union syntax requires python3.10 or higher",
)
def test_union_catchall():
    @ovld
    def f(t: UnionTypes):
        return "yes"

    @ovld
    def f(t: type[object]):
        return "no"

    assert f(typing.Optional[float]) == "yes"
    assert f(typing.Union[int, str]) == "yes"
    assert f(int | str) == "yes"
    assert f(int) == "no"


def test_annotated():
    @ovld
    def f(x: typing.Annotated[int, "blah"]):
        return x * x

    assert f(7) == 49


def test_iterable():
    @ovld
    def f(xs: typing.Iterable):
        return sum(xs)

    assert f([1, 2, 3]) == 6
    assert f(range(4)) == 6


def test_plain_type_argument():
    @ovld
    def f(t: type[list]):
        return "list"

    @f.register
    def f(t: type):
        return "other type"

    @f.register
    def f(t: object):
        return "other anything"

    assert f(list) == "list"
    assert f(list[int]) == "list"
    assert f(dict) == "other type"
    assert f(1234) == "other anything"


def test_incidental_type_argument():
    @ovld
    def f(t: object):
        return "ok"

    @ovld
    def f(t: type[str]):
        return "hmmm"

    assert f(list) == "ok"


def test_parametrized():
    @ovld
    def f(t: type[list[int]]):
        return "list of int"

    @f.register
    def f(t: type[list[str]]):
        return "list of str"

    @f.register
    def f(t: type[list]):
        return "list"

    assert f(list[int]) == "list of int"
    assert f(list[str]) == "list of str"
    assert f(list[object]) == "list"
    assert f(list) == "list"


def test_parametrized_protocols():
    @ovld
    def f(t: type[typing.Iterable[object]]):
        return "iterable"

    # @f.register
    # def f(t: type[typing.Iterable[Animal]]):
    #     return "iterable of animals"

    assert f(list[int]) == "iterable"
    # assert f(list[Animal]) == "iterable of animals"
    # assert f(list[Mammal]) == "iterable of animals"


@pytest.mark.skipif(
    sys.version_info < (3, 12),
    reason="Generic type syntax requires python3.12 or higher",
)
def test_parametrized_class():
    code = """class Thing[T]: ..."""
    exec(code, globals(), globals())

    Thing = globals()["Thing"]

    @ovld
    def f(t: type[Thing[Animal]]):
        return "thingy"

    assert f(Thing[Mammal]) == "thingy"


class Booboo:
    pass


def test_string_annotation():
    @ovld
    def f(t: "int"):
        return "int"

    @f.register
    def f(t: "list"):
        return "list"

    @f.register
    def f(t: "Booboo"):
        return "boo"

    assert f(1234) == "int"
    assert f([1, 2, 3, 4]) == "list"
    assert f(Booboo()) == "boo"


def test_ovld_outside_scope():
    def indirect():
        def f(x: int):
            return x * x

        return f

    with pytest.raises(UsageError):
        ovld(indirect())

    o = ovld(indirect(), fresh=True)
    assert o(3) == 9


def test_display(capsys, file_regression):
    @ovld
    def f(x: Dependent[int, Equals(0)]):
        return "zero"

    @f.register
    def f(x: Dependent[int, Equals(1)]):
        return "one"

    @f.register
    def f(x: Dependent[str, StartsWith("hell")]):
        return "H"

    @f.register
    def f(x: Dependent[str, StartsWith("hello")]):
        return "H"

    @f.register
    def f(x: int):
        return "yes"

    @f.register
    def f(x: str):
        return "yes"

    @f.register
    def f(x: object):
        return "done"

    @f.register(priority=10)
    def f(x: object):
        return "start"

    @f.register
    def f(x: object, *, k: str):
        return "keyy"

    f.display_methods()
    for arg in (13, 0, 1, "hello"):
        print("=" * 80)
        print(f"Resolve f({arg!r})")
        print("=" * 80)
        f.display_resolution(arg)

    print("=" * 80)
    f.display_resolution(123, k="oi")

    out, err = capsys.readouterr()
    assert not err
    out = out.replace(__file__, "THIS_FILE")

    def renumber(m):
        return "THIS_FILE:" + str(int(m[1]) - test_display.__code__.co_firstlineno)

    out = re.sub(string=out, pattern=r"THIS_FILE:([0-9]+)", repl=renumber)
    file_regression.check(out)


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="union syntax requires python3.10 or higher",
)
def test_display_more(capsys, file_regression):
    @ovld
    def f(x: int | str):
        return 0

    @ovld
    def f(x: int & Equals[0]):
        return 1

    @ovld
    def f(x: dict[str, int]):
        return 2

    @ovld
    def f(x: tuple[list[int], set[float]]):
        return 3

    f.display_methods()

    out, err = capsys.readouterr()
    assert not err
    out = out.replace(__file__, "THIS_FILE")

    def renumber(m):
        return "THIS_FILE:" + str(int(m[1]) - test_display_more.__code__.co_firstlineno)

    out = re.sub(string=out, pattern=r"THIS_FILE:([0-9]+)", repl=renumber)
    file_regression.check(out)


def test_closure():
    def make(n):
        @ovld
        def f(xs: list):
            return [recurse(x) for x in xs]

        @ovld
        def f(x: int):
            return x * n

        return f

    twice = make(2)
    thrice = make(3)

    assert twice([1, 2, 3]) == [2, 4, 6]
    assert thrice([1, 2, 3]) == [3, 6, 9]


def test_closure_plus_recurse():
    def make(n):
        @ovld
        def f(xs: list):
            return [recurse(x, n) for x in xs]

        @ovld
        def f(x: int, y: int):
            return x * y

        return f

    twice = make(2)
    thrice = make(3)

    assert twice([1, 2, 3]) == [2, 4, 6]
    assert thrice([1, 2, 3]) == [3, 6, 9]


def test_no_kwargs():
    with pytest.raises(TypeError):

        @ovld
        def f(x: int, **kw):
            return x

        f(123)


def test_inconsistent_position_naming():
    @ovld
    def f(x: int, y: int):
        return x + y

    @ovld
    def f(y: int):
        return y * y

    with pytest.raises(
        TypeError, match="declared in different positions by different methods"
    ):
        f(123)


def test_inconsistent_naming():
    @ovld
    def f(x: int, y: int):
        return x + y

    @ovld
    def f(x: str, *, y: str):
        return y + y

    with pytest.raises(
        TypeError,
        match="declared in a positional and keyword setting by different methods",
    ):
        f(1, 2)


def test_uniform_positional_names():
    @ovld
    def f(x: int, y: int):
        return x + y

    @ovld
    def f(x: int):
        return x * x

    assert f(3, 4) == 7
    assert f(3) == 9

    assert f(x=3, y=4) == 7
    assert f(x=3) == 9


def test_strict_positional():
    @ovld
    def f(x: int, y: int, /):
        return x + y

    @ovld
    def f(y: int, /):
        return y * y

    assert f(3, 4) == 7
    assert f(3) == 9
    with pytest.raises(TypeError):
        f(x=3, y=4)


def test_keywords():
    @ovld
    def f(name: str, *, hello: int):
        return "hello" * hello + " " + name

    @ovld
    def f(name: str, *, hello: str):
        return hello + " " + name

    @f.register
    def f(name: str, *, goodbye: int):
        return "goodbye" * goodbye + " " + name

    assert f("Helena", hello=3) == "hellohellohello Helena"
    assert f("Helena", hello="Bonjour") == "Bonjour Helena"
    assert f("Gertrude", goodbye=2) == "goodbyegoodbye Gertrude"


def test_keywords_recurse():
    @ovld
    def f(xs: list, *, factor: int):
        return [recurse(x, factor=factor) for x in xs]

    @ovld
    def f(x: int, *, factor: int):
        return x * factor

    assert f([1, 2, 3], factor=3) == [3, 6, 9]


def test_resolve_all():
    @ovld
    def f(x: str):
        return "str"

    @ovld(priority=7)
    def f(x: str):
        return "MAX"

    @ovld(priority=1)
    def f(x: Regexp["^hello"]):
        return "hello"

    @ovld
    def f(x: Regexp["world$"]):
        return "world"

    @ovld
    def f(x: int):
        return "int"

    assert [t() for t in f.resolve_all("hello world")] == ["MAX", "hello", "world", "str"]
    assert [t() for t in f.resolve_all("hello, you!")] == ["MAX", "hello", "str"]
    assert [t() for t in f.resolve_all(73)] == ["int"]
    assert [t() for t in f.resolve_all(object())] == []


def test_passing_types_to_normal_func():
    @ovld
    def f(x):
        return x

    assert f(list[int]) == list[int]


def test_doc(file_regression):
    @ovld
    def mushroom(x: int):
        """Unlike mushrooms, this function doesn't do anything."""
        return None

    @ovld
    def mushroom(x: str):
        """Naming mushrooms."""
        return None

    @ovld
    def mushroom(x: str, y: object, *, beauty, bigness):
        """Whatever."""
        return None

    doc = f"{mushroom.__name__}{inspect.signature(mushroom)}\n\n"
    doc += inspect.getdoc(mushroom)

    file_regression.check(doc)

    assert inspect.getdoc(mushroom) == inspect.getdoc(mushroom.__ovld__)
    assert inspect.signature(mushroom) == inspect.signature(mushroom.__ovld__)


def test_doc2(file_regression):
    @ovld
    def mushroom(x: int):
        """Unlike mushrooms, this function doesn't do anything."""
        return None

    @ovld
    def mushroom(x: str):
        return None

    @ovld
    def mushroom(x: str, y: object, *, beauty, bigness):
        return None

    doc = inspect.getdoc(mushroom)
    doc = f"{mushroom.__name__}{inspect.signature(mushroom)}\n\n" + doc

    file_regression.check(doc)


def test_method_doc(file_regression):
    class Mushroom:
        @ovld
        def rise(self, x: int):
            """Unlike mushrooms, this function doesn't do anything."""
            return None

        @ovld
        def rise(self, x: str):
            return None

        @ovld
        def rise(self, x: str, y: object, *, beauty, bigness):
            return None

    mushroom = Mushroom()
    doc = inspect.getdoc(mushroom.rise)
    doc = f"{mushroom.rise.__name__}{inspect.signature(mushroom.rise)}\n\n" + doc

    file_regression.check(doc)


def test_possible_clobbering():
    @ovld
    def f(type: typing.Literal[0]):
        return 0

    @ovld
    def f(type: typing.Literal[1]):
        return 1

    assert f(0) == 0
    assert f(1) == 1



from typing import Literal

import pytest

from .common import (
    function_builder,
    multimethod_dispatch,
    ovld_dispatch,
    plum_dispatch,
    with_functions,
)

###############################
# multiple dispatch libraries #
###############################


@function_builder
def make_fib(dispatch):
    @dispatch
    def fib(n: Literal[0]):
        return 0

    @dispatch
    def fib(n: Literal[1]):
        return 1

    @dispatch
    def fib(n: int):
        return fib(n - 1) + fib(n - 2)

    return fib


##########
# normal #
##########


def fib_normal(n):
    if n <= 1:
        return n
    else:
        return fib_normal(n - 1) + fib_normal(n - 2)


####################
# Test definitions #
####################


@pytest.mark.benchmark(group="fib")
@with_functions(
    ovld=make_fib(ovld_dispatch),
    plum=make_fib(plum_dispatch),
    multimethod=make_fib(multimethod_dispatch),
    custom=fib_normal,
)
def test_fib(fn, benchmark):
    result = benchmark(fn, 8)
    assert result == 21



import inspect
import math
from dataclasses import dataclass
from functools import partial
from itertools import count
from types import CodeType

from .mro import sort_types
from .recode import generate_dependent_dispatch
from .utils import MISSING, CodegenInProgress, is_dependent, subtler_type


class TypeMap(dict):
    """Represents a mapping from types to handlers.

    The mro of a type is considered when getting the handler, so setting the
    [object] key creates a default for all objects.

    typemap[some_type] returns a tuple of a handler and a "level" that
    represents the distance from the handler to the type `object`. Essentially,
    the level is the index of the type for which the handler was registered
    in the mro of `some_type`. So for example, `object` has level 0, a class
    that inherits directly from `object` has level 1, and so on.
    """

    def __init__(self):
        self.entries = {}
        self.types = set()

    def register(self, obj_t, handler):
        """Register a handler for the given object type."""
        self.clear()
        self.types.add(obj_t)
        s = self.entries.setdefault(obj_t, set())
        s.add(handler)

    def __missing__(self, obj_t):
        """Get the handler for the given type.

        The result is cached so that the normal dict getitem will find it
        the next time getitem is called.
        """
        results = {}
        groups = list(sort_types(obj_t, self.types))

        for lvl, grp in enumerate(reversed(groups)):
            for cls in grp:
                handlers = self.entries.get(cls, None)
                if handlers:
                    results.update({h: lvl for h in handlers})

        if results:
            self[obj_t] = results
            return results
        else:
            raise KeyError(obj_t)


@dataclass
class Candidate:
    handler: object
    base_handler: object
    priority: float
    specificity: tuple
    tiebreak: int

    def sort_key(self):
        return self.priority, sum(self.specificity), self.tiebreak

    def dominates(self, other):
        if self.priority > other.priority:
            return True
        elif self.specificity != other.specificity:
            return all(s1 >= s2 for s1, s2 in zip(self.specificity, other.specificity))
        else:
            return self.tiebreak > other.tiebreak


class MultiTypeMap(dict):
    """Represents a mapping from tuples of types to handlers.

    The mro is taken into account to find a match. If multiple registered
    handlers match the tuple of types that's given, if one of the handlers is
    more specific than every other handler, that handler is returned.
    Otherwise, the resolution is considered ambiguous and an error is raised.

    Handler A, registered for types (A1, A2, ..., An), is more specific than
    handler B, registered for types (B1, B2, ..., Bn), if there exists n such
    that An is more specific than Bn, and for all n, either An == Bn or An is
    more specific than Bn. An is more specific than Bn if An is a direct or
    indirect subclass of Bn.

    In other words, [int, object] is more specific than [object, object] and
    less specific than [int, int], but it is neither less specific nor more
    specific than [object, int] (which means there is an ambiguity).
    """

    def __init__(self, name="_ovld", key_error=KeyError, ovld=None):
        self.maps = {}
        self.priorities = {}
        self.tiebreaks = {}
        self.dependent = {}
        self.type_tuples = {}
        self.empty = MISSING
        self.key_error = key_error
        self.name = name
        self.dispatch_id = count()
        self.all = {}
        self.errors = {}
        self.ovld = ovld
        self.in_progress = set()

    def mro(self, obj_t_tup, specialize=True):
        specificities = {}
        candidates = None
        nargs = len([t for t in obj_t_tup if not isinstance(t, tuple)])
        names = {t[0] for t in obj_t_tup if isinstance(t, tuple)}

        for i, cls in enumerate(obj_t_tup):
            if isinstance(cls, tuple):
                i, cls = cls

            try:
                results = self.maps[i][cls]
            except KeyError:
                results = {}

            results = {
                handler: spc
                for (handler, sig), spc in results.items()
                if sig.req_pos <= nargs <= (math.inf if sig.vararg else sig.max_pos)
                and not (sig.req_names - names)
            }

            try:
                vararg_results = self.maps[-1][cls]
            except KeyError:
                vararg_results = {}

            vararg_results = {
                handler: spc
                for (handler, sig), spc in vararg_results.items()
                if sig.req_pos <= nargs and i >= sig.max_pos
            }

            results.update(vararg_results)

            if candidates is None:
                candidates = set(results.keys())
            else:
                candidates &= results.keys()
            for c in candidates:
                specificities.setdefault(c, []).append(results[c])

        def get_handler(func):
            if specialize and (specializer := getattr(func, "specializer", False)):
                key = (func, obj_t_tup)
                if key in self.in_progress:
                    raise CodegenInProgress()
                self.in_progress.add(key)
                return specializer(self, func, obj_t_tup)
            else:
                return func

        candidates = [
            Candidate(
                handler=h,
                base_handler=c,
                priority=self.priorities.get(c, 0),
                specificity=tuple(specificities[c]),
                tiebreak=self.tiebreaks.get(c, 0),
            )
            for c in candidates
            if (h := get_handler(c))
        ]

        # The sort ensures that if candidate A dominates candidate B, A will
        # appear before B in the list. That's because it must dominate all
        # other possibilities on all arguments, so the sum of all specificities
        # has to be greater.
        # Note: priority is always more important than specificity

        candidates.sort(key=Candidate.sort_key, reverse=True)

        self.all[obj_t_tup] = {getattr(c.handler, "__code__", None) for c in candidates}

        processed = set()

        def _pull(candidates):
            candidates = [c for c in candidates if c.handler not in processed]
            if not candidates:
                return
            rval = [candidates[0]]
            c1 = candidates[0]
            for c2 in candidates[1:]:
                if c1.dominates(c2):
                    # Candidate 1 dominates candidate 2
                    continue
                else:
                    processed.add(c2.handler)
                    # Candidate 1 does not dominate candidate 2, so we add it
                    # to the list.
                    rval.append(c2)
            yield rval
            if len(rval) >= 1:
                yield from _pull(candidates[1:])

        return list(_pull(candidates))

    def register(self, sig, handler):
        """Register a handler for a tuple of argument types.

        Arguments:
            sig: A Signature object.
            handler: A function to handle the tuple.
        """
        self.clear()

        obj_t_tup = sig.types
        entry = (handler, sig)
        if not obj_t_tup:
            self.empty = entry

        self.priorities[handler] = sig.priority
        self.tiebreaks[handler] = sig.tiebreak
        self.type_tuples[handler] = obj_t_tup
        self.dependent[handler] = any(
            is_dependent(t[1] if isinstance(t, tuple) else t) for t in obj_t_tup
        )

        for i, cls in enumerate(obj_t_tup):
            if isinstance(cls, tuple):
                i, cls = cls
            if i not in self.maps:
                self.maps[i] = TypeMap()
            self.maps[i].register(cls, entry)

        if sig.vararg:  # pragma: no cover
            # TODO: either add this back in, or remove it
            if -1 not in self.maps:
                self.maps[-1] = TypeMap()
            self.maps[-1].register(object, entry)

    def display_methods(self):
        for h, prio in sorted(self.priorities.items(), key=lambda kv: kv[1], reverse=True):
            prio = ":".join(map(str, prio))
            prio = f"[{prio}]"
            width = 6
            print(f"{prio:{width}} \033[1m{h.__name__}\033[0m")
            co = h.__code__
            print(f"{'':{width - 2}} @ {co.co_filename}:{co.co_firstlineno}")

    def _resolve_all_helper(self, *args, **kwargs):
        def dependent_match(tup, args):
            for t, a in zip(tup, args):
                if isinstance(t, tuple):
                    t = t[1]
                    a = a[1]
                if is_dependent(t) and not isinstance(a, t):
                    return False
            return True

        argt = [
            *map(subtler_type, args),
            *[(k, subtler_type(v)) for k, v in kwargs.items()],
        ]
        for grp in self.mro(tuple(argt)):
            yield [
                (
                    c,
                    dependent_match(
                        self.type_tuples[c.base_handler], [*args, *kwargs.items()]
                    ),
                )
                for c in grp
            ]

    def resolve_all(self, *args, **kwargs):
        for grp in self._resolve_all_helper(*args, **kwargs):
            for c, m in grp:
                if m:
                    yield partial(c.handler, *args, **kwargs)

    def display_resolution(self, *args, **kwargs):
        message = "No method will be called."
        finished = False
        rank = 1
        for grp in self._resolve_all_helper(*args, **kwargs):
            grp.sort(key=lambda x: x[0].handler.__name__)
            ambiguous = len([m for _, m in grp if m]) > 1
            for c, m in grp:
                handler = c.handler
                color = "\033[0m"
                if finished:
                    bullet = "--"
                    color = "\033[1;90m"
                elif not m:
                    bullet = "!="
                    color = "\033[1;90m"
                elif ambiguous:
                    bullet = "=="
                    color = "\033[1;31m"
                else:
                    bullet = f"#{rank}"
                    if rank == 1:
                        message = f"{handler.__name__} will be called first."
                        color = "\033[1;32m"
                    rank += 1
                spec = ".".join(map(str, c.specificity))
                prios = ":".join(map(str, c.priority))
                lvl = f"[{prios}:{spec}]"
                width = 2 * len(args) + 6
                print(f"{color}{bullet} {lvl:{width}} {handler.__name__}")
                co = handler.__code__
                print(f"   {'':{width - 1}}@ {co.co_filename}:{co.co_firstlineno}\033[0m")
            if ambiguous:
                message += (
                    " There is ambiguity between multiple matching methods, marked '=='."
                )
                finished = True
        print("Resolution:", message)

    def wrap_dependent(self, tup, group, next_call):
        htup = [(c.handler, self.type_tuples[c.base_handler]) for c in group]
        slf = "self, " if inspect.getfullargspec(group[0].handler).args[0] == "self" else ""
        return generate_dependent_dispatch(
            tup,
            htup,
            next_call,
            slf,
            name=f"{self.name}.specialized_dispatch_{next(self.dispatch_id)}",
            err=self.key_error(tup, group),
            nerr=self.key_error(tup, ()),
        )

    def resolve(self, obj_t_tup):
        results = self.mro(obj_t_tup)
        if not results:
            raise self.key_error(obj_t_tup, ())

        funcs = []
        for group in reversed(results):
            handlers = [c.handler for c in group]
            dependent = any(self.dependent[c.base_handler] for c in group)
            if dependent:
                nxt = self.wrap_dependent(obj_t_tup, group, funcs[-1] if funcs else None)
            elif len(group) != 1:
                nxt = None
            else:
                nxt = handlers[0]
            codes = [h.__code__ for h in handlers if hasattr(h, "__code__")]
            funcs.append((nxt, codes))

        funcs.reverse()

        parents = []
        for group, (func, codes) in zip(results, funcs):
            tups = [obj_t_tup] if not parents else [(parent, *obj_t_tup) for parent in parents]
            if func is None:
                for tup in tups:
                    self.errors[tup] = self.key_error(obj_t_tup, group)
                break
            else:
                for tup in tups:
                    self[tup] = func
            if not codes:
                break
            parents = codes

        return True

    def __missing__(self, obj_t_tup):
        if obj_t_tup and isinstance(obj_t_tup[0], CodeType):
            real_tup = obj_t_tup[1:]
            self[real_tup]
            if obj_t_tup[0] not in self.all[real_tup]:
                return self[real_tup]
            elif obj_t_tup in self.errors:
                raise self.errors[obj_t_tup]
            elif obj_t_tup in self:  # pragma: no cover
                # PROBABLY not reachable
                return self[obj_t_tup]
            else:
                raise self.key_error(real_tup, ())

        if not obj_t_tup:
            if self.empty is MISSING:  # pragma: no cover
                # Might not be reachable because of codegen
                raise self.key_error(obj_t_tup, ())
            else:
                return self.empty[0]

        self.resolve(obj_t_tup)
        if obj_t_tup in self.errors:
            raise self.errors[obj_t_tup]
        else:
            return self[obj_t_tup]

    def __call__(self, *obj_t_tup, after=None):
        if after:
            return self[(after, *obj_t_tup)]
        else:
            return self[obj_t_tup]



import sys
from dataclasses import dataclass, fields
from numbers import Number
from typing import Literal, Union

import pytest

from ovld import ovld, recurse
from ovld.types import Dataclass


@ovld
def calc(num: Number):
    return num


@ovld
def calc(tup: tuple):
    return calc(*tup)


@ovld
def calc(op: Literal["add"], x: object, y: object):
    return calc(x) + calc(y)


@ovld
def calc(op: Literal["sub"], x: object, y: object):
    return calc(x) - calc(y)


@ovld
def calc(op: Literal["mul"], x: object, y: object):
    return calc(x) * calc(y)


@ovld
def calc(op: Literal["div"], x: object, y: object):
    return calc(x) / calc(y)


@ovld
def calc(op: Literal["pow"], x: object, y: object):
    return calc(x) ** calc(y)


@ovld
def calc(op: Literal["sqrt"], x: object):
    return calc(x) ** 0.5


def test_calc():
    expr = (
        "add",
        ("mul", ("sqrt", 4), 7),
        ("div", ("add", 6, 4), ("sub", 5, 3)),
    )
    expected_result = 19
    assert calc(expr) == expected_result


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


@ovld
def deserialize(t: type[Dataclass], data: dict):
    kwargs = {f.name: recurse(f.type, data[f.name]) for f in fields(t) if f.name in data}
    return t(**kwargs)


@ovld
def deserialize(t: type[list], data: list):
    (lt,) = t.__args__
    return [recurse(lt, x) for x in data]


@ovld
def deserialize(t: type[dict], data: dict):
    (
        kt,
        vt,
    ) = t.__args__
    return {recurse(kt, k): recurse(vt, v) for k, v in data.items()}


@ovld
def deserialize(t: type[Union[int, str]], data: Union[int, str]):
    assert isinstance(data, t)
    return data


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="doesn't work, not worth fixing",
)
def test_deserialize():
    data = {
        "countries": {
            "canada": {
                "languages": ["English", "French"],
                "capital": "Ottawa",
                "population": 39_000_000,
                "citizens": [
                    {
                        "name": "Olivier",
                        "birthyear": 1985,
                        "hometown": "Montreal",
                    },
                    {
                        "name": "Abraham",
                        "birthyear": 2018,
                        "hometown": "Shawinigan",
                    },
                ],
            }
        }
    }

    expected = World(
        countries={
            "canada": Country(
                languages=["English", "French"],
                capital="Ottawa",
                population=39_000_000,
                citizens=[
                    Citizen(
                        name="Olivier",
                        birthyear=1985,
                        hometown="Montreal",
                    ),
                    Citizen(
                        name="Abraham",
                        birthyear=2018,
                        hometown="Shawinigan",
                    ),
                ],
            )
        }
    )

    assert deserialize(World, data) == expected



import json

from pytest_benchmark import utils

from ovld import ovld, recurse


@ovld
def _cleanup(obj: str | int | float):
    return obj


@ovld
def _cleanup(obj: list):
    return [recurse(x) for x in obj]


@ovld
def _cleanup(obj: dict):
    if type(obj) is not dict:
        return f"UNSERIALIZABLE[{obj}]"
    return {k: recurse(v) for k, v in obj.items()}


@ovld
def _cleanup(obj: object):
    return f"UNSERIALIZABLE[{obj}]"


def safer_dumps(obj, **kwargs):
    # multimethod is not safe for dump of benchmarks because it's a subclass of dict
    # with non-str keys and the json serialization just craps out.
    return json.dumps(_cleanup(obj), **kwargs)


utils._cleanup = _cleanup
utils.safe_dumps.__code__ = safer_dumps.__code__



from functools import singledispatchmethod

import pytest
from multipledispatch import dispatch as multipledispatch_dispatch

from ovld import recurse
from ovld.core import OvldBase

from .common import (
    function_builder,
    multimethod_dispatch,
    ovld_dispatch,
    plum_dispatch,
    runtype_dispatch,
    with_functions,
)


class BaseMulter:
    def __init__(self, factor):
        self.factor = factor


@function_builder
def make_multer(dispatch):
    class Multer(BaseMulter):
        @dispatch
        def __call__(self, x: list):
            return [self(a) for a in x]

        @dispatch
        def __call__(self, x: tuple):
            return tuple(self(a) for a in x)

        @dispatch
        def __call__(self, x: dict):
            return {k: self(v) for k, v in x.items()}

        @dispatch
        def __call__(self, x: object):
            return x * self.factor

    return Multer


#####################
# ovld with recurse #
#####################


class OvldRecurseMulter(BaseMulter, OvldBase):
    def __call__(self, x: list):
        return [recurse(a) for a in x]

    def __call__(self, x: tuple):
        return tuple(recurse(a) for a in x)

    def __call__(self, x: dict):
        return {k: recurse(v) for k, v in x.items()}

    def __call__(self, x: object):
        return x * self.factor


####################
# multipledispatch #
####################


# shim doesn't work


class MultipleDispatchMulter(BaseMulter):
    @multipledispatch_dispatch(list)
    def __call__(self, x: list):
        return [self(a) for a in x]

    @multipledispatch_dispatch(tuple)
    def __call__(self, x: tuple):
        return tuple(self(a) for a in x)

    @multipledispatch_dispatch(dict)
    def __call__(self, x: dict):
        return {k: self(v) for k, v in x.items()}

    @multipledispatch_dispatch(object)
    def __call__(self, x: object):
        return x * self.factor


##################
# singledispatch #
##################


class SingleDispatchMulter(BaseMulter):
    @singledispatchmethod
    def __call__(self, x):
        return x * self.factor

    @__call__.register
    def _(self, x: tuple):
        return tuple(self(a) for a in x)

    @__call__.register
    def _(self, x: dict):
        return {k: self(v) for k, v in x.items()}

    @__call__.register
    def _(self, x: list):
        return [self(a) for a in x]


##############
# isinstance #
##############


class IsinstanceMulter:
    def __init__(self, factor):
        self.factor = factor

    def __call__(self, x):
        if isinstance(x, dict):
            return {k: self(v) for k, v in x.items()}
        elif isinstance(x, tuple):
            return tuple(self(a) for a in x)
        elif isinstance(x, list):
            return [self(a) for a in x]
        else:
            return x * self.factor


####################
# Test definitions #
####################


A = {"xs": list(range(0, 50)), "ys": ("o", (6, 7))}
C = {"xs": list(range(0, 150, 3)), "ys": ("ooo", (18, 21))}


@pytest.mark.benchmark(group="multer")
@with_functions(
    ovld=make_multer(ovld_dispatch),
    ovld__recurse=OvldRecurseMulter,
    plum=make_multer(plum_dispatch),
    multimethod=make_multer(multimethod_dispatch),
    singledispatch=SingleDispatchMulter,
    multipledispatch=MultipleDispatchMulter,
    runtype=make_multer(runtype_dispatch),
    custom=IsinstanceMulter,
)
def test_multer(fn, benchmark):
    result = benchmark(fn(3), A)
    assert result == C
